# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import re

from .. import _backend_config
from .._errors import pretty_message
from .._ffi import get_library, register_ffi
from ..errors import LibraryNotFoundError

from cffi import FFI


__all__ = [
    'is_libressl',
    'libcrypto',
    'libressl_version',
    'libressl_version_info',
    'version',
    'version_info',
]

libcrypto_path = _backend_config().get('libcrypto_path')
if libcrypto_path is None:
    libcrypto_path = get_library('crypto', 'libcrypto.dylib', '42')
if not libcrypto_path:
    raise LibraryNotFoundError('The library libcrypto could not be found')

try:
    vffi = FFI()
    vffi.cdef("const char *SSLeay_version(int type);")
    version_string = vffi.string(vffi.dlopen(libcrypto_path).SSLeay_version(0)).decode('utf-8')
except (AttributeError):
    vffi = FFI()
    vffi.cdef("const char *OpenSSL_version(int type);")
    version_string = vffi.string(vffi.dlopen(libcrypto_path).OpenSSL_version(0)).decode('utf-8')

is_libressl = 'LibreSSL' in version_string

version_match = re.search('\\b(\\d\\.\\d\\.\\d[a-z]*)\\b', version_string)
if not version_match:
    version_match = re.search('(?<=LibreSSL )(\\d\\.\\d(\\.\\d)?)\\b', version_string)
if not version_match:
    raise LibraryNotFoundError('Error detecting the version of libcrypto')
version = version_match.group(1)
version_parts = re.sub('(\\d)([a-z]+)', '\\1.\\2', version).split('.')
version_info = tuple(int(part) if part.isdigit() else part for part in version_parts)

# LibreSSL is compatible with libcrypto from OpenSSL 1.0.1
libressl_version = ''
libressl_version_info = tuple()
if is_libressl:
    libressl_version = version
    libressl_version_info = version_info
    version = '1.0.1'
    version_info = (1, 0, 1)

ffi = FFI()

libcrypto = ffi.dlopen(libcrypto_path)
register_ffi(libcrypto, ffi)

if version_info < (0, 9, 8):
    raise LibraryNotFoundError(pretty_message(
        '''
        OpenSSL versions older than 0.9.8 are not supported - found version %s
        ''',
        version
    ))

if version_info < (1, 1):
    ffi.cdef("""
        void ERR_load_crypto_strings(void);
        void ERR_free_strings(void);
    """)


if version_info >= (3, ):
    ffi.cdef("""
        typedef ... OSSL_LIB_CTX;
        typedef ... OSSL_PROVIDER;

        int OSSL_PROVIDER_available(OSSL_LIB_CTX *libctx, const char *name);
        OSSL_PROVIDER *OSSL_PROVIDER_load(OSSL_LIB_CTX *libctx, const char *name);
    """)

# The typedef uintptr_t lines here allow us to check for a NULL pointer,
# without having to redefine the structs in our code. This is kind of a hack,
# but it should cause problems since we treat these as opaque.
ffi.cdef("""
    typedef ... EVP_MD;
    typedef uintptr_t EVP_CIPHER_CTX;
    typedef ... EVP_CIPHER;
    typedef ... ENGINE;
    typedef uintptr_t EVP_PKEY;
    typedef uintptr_t X509;
    typedef uintptr_t DH;
    typedef uintptr_t RSA;
    typedef uintptr_t DSA;
    typedef uintptr_t EC_KEY;
    typedef ... EVP_MD_CTX;
    typedef ... EVP_PKEY_CTX;
    typedef ... BN_GENCB;
    typedef ... BIGNUM;

    unsigned long ERR_get_error(void);
    char *ERR_error_string(unsigned long e, char *buf);
    unsigned long ERR_peek_error(void);

    void OPENSSL_config(const char *config_name);

    EVP_CIPHER_CTX *EVP_CIPHER_CTX_new(void);
    void EVP_CIPHER_CTX_free(EVP_CIPHER_CTX *ctx);

    int EVP_CIPHER_CTX_set_key_length(EVP_CIPHER_CTX *x, int keylen);
    int EVP_CIPHER_CTX_set_padding(EVP_CIPHER_CTX *x, int padding);
    int EVP_CIPHER_CTX_ctrl(EVP_CIPHER_CTX *ctx, int type, int arg, void *ptr);

    const EVP_CIPHER *EVP_aes_128_cbc(void);
    const EVP_CIPHER *EVP_aes_192_cbc(void);
    const EVP_CIPHER *EVP_aes_256_cbc(void);
    const EVP_CIPHER *EVP_des_cbc(void);
    const EVP_CIPHER *EVP_des_ede_cbc(void);
    const EVP_CIPHER *EVP_des_ede3_cbc(void);
    const EVP_CIPHER *EVP_rc4(void);
    const EVP_CIPHER *EVP_rc2_cbc(void);

    int EVP_EncryptInit_ex(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher,
                    ENGINE *impl, const char *key,
                    const char *iv);
    int EVP_EncryptUpdate(EVP_CIPHER_CTX *ctx, char *out, int *outl,
                    const char *in, int inl);
    int EVP_EncryptFinal_ex(EVP_CIPHER_CTX *ctx, char *out, int *outl);

    int EVP_DecryptInit_ex(EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher,
                    ENGINE *impl, const char *key,
                    const char *iv);
    int EVP_DecryptUpdate(EVP_CIPHER_CTX *ctx, char *out, int *outl,
                    const char *in, int inl);
    int EVP_DecryptFinal_ex(EVP_CIPHER_CTX *ctx, char *out, int *outl);

    EVP_PKEY *d2i_AutoPrivateKey(EVP_PKEY **a, const char **pp,
                    long length);
    EVP_PKEY *d2i_PUBKEY(EVP_PKEY **a, const char **pp, long length);
    int i2d_PUBKEY(EVP_PKEY *a, char **pp);
    void EVP_PKEY_free(EVP_PKEY *key);

    X509 *d2i_X509(X509 **px, const char **in, int len);
    int i2d_X509(X509 *x, char **out);
    EVP_PKEY *X509_get_pubkey(X509 *x);
    void X509_free(X509 *a);

    RSA *EVP_PKEY_get1_RSA(EVP_PKEY *pkey);
    void RSA_free(RSA *r);

    int RSA_public_encrypt(int flen, const char *from,
                    char *to, RSA *rsa, int padding);
    int RSA_private_encrypt(int flen, const char *from,
                    char *to, RSA *rsa, int padding);
    int RSA_public_decrypt(int flen, const char *from,
                    char *to, RSA *rsa, int padding);
    int RSA_private_decrypt(int flen, const char *from,
                    char *to, RSA *rsa, int padding);

    int EVP_DigestUpdate(EVP_MD_CTX *ctx, const void *d, unsigned int cnt);

    const EVP_MD *EVP_md5(void);
    const EVP_MD *EVP_sha1(void);
    const EVP_MD *EVP_sha224(void);
    const EVP_MD *EVP_sha256(void);
    const EVP_MD *EVP_sha384(void);
    const EVP_MD *EVP_sha512(void);

    int PKCS12_key_gen_uni(char *pass, int passlen, char *salt,
                    int saltlen, int id, int iter, int n,
                    char *out, const EVP_MD *md_type);

    void BN_free(BIGNUM *a);
    int BN_dec2bn(BIGNUM **a, const char *str);

    DH *DH_new(void);
    int DH_generate_parameters_ex(DH *dh, int prime_len, int generator, BN_GENCB *cb);
    int i2d_DHparams(const DH *a, char **pp);
    void DH_free(DH *dh);

    RSA *RSA_new(void);
    int RSA_generate_key_ex(RSA *rsa, int bits, BIGNUM *e, BN_GENCB *cb);
    int i2d_RSAPublicKey(RSA *a, char **pp);
    int i2d_RSAPrivateKey(RSA *a, char **pp);

    DSA *DSA_new(void);
    int DSA_generate_parameters_ex(DSA *dsa, int bits,
                    const char *seed, int seed_len, int *counter_ret,
                    unsigned long *h_ret, BN_GENCB *cb);
    int DSA_generate_key(DSA *a);
    int i2d_DSA_PUBKEY(const DSA *a, char **pp);
    int i2d_DSAPrivateKey(const DSA *a, char **pp);
    void DSA_free(DSA *dsa);

    EC_KEY *EC_KEY_new_by_curve_name(int nid);
    int EC_KEY_generate_key(EC_KEY *key);
    void EC_KEY_set_asn1_flag(EC_KEY *, int);
    int i2d_ECPrivateKey(EC_KEY *key, char **out);
    int i2o_ECPublicKey(EC_KEY *key, char **out);
    void EC_KEY_free(EC_KEY *key);
""")

if version_info < (3, ):
    ffi.cdef("""
        int EVP_PKEY_size(EVP_PKEY *pkey);
    """)
else:
    ffi.cdef("""
        int EVP_PKEY_get_size(EVP_PKEY *pkey);
    """)

if version_info < (1, 1):
    ffi.cdef("""
        EVP_MD_CTX *EVP_MD_CTX_create(void);
        void EVP_MD_CTX_destroy(EVP_MD_CTX *ctx);
    """)
else:
    ffi.cdef("""
        EVP_MD_CTX *EVP_MD_CTX_new(void);
        void EVP_MD_CTX_free(EVP_MD_CTX *ctx);
    """)

if version_info < (1,):
    ffi.cdef("""
        typedef ... *DSA_SIG;
        typedef ... *ECDSA_SIG;

        DSA_SIG *DSA_do_sign(const char *dgst, int dlen, DSA *dsa);
        ECDSA_SIG *ECDSA_do_sign(const char *dgst, int dgst_len, EC_KEY *eckey);

        DSA_SIG *d2i_DSA_SIG(DSA_SIG **v, const char **pp, long length);
        ECDSA_SIG *d2i_ECDSA_SIG(ECDSA_SIG **v, const char **pp, long len);

        int i2d_DSA_SIG(const DSA_SIG *a, char **pp);
        int i2d_ECDSA_SIG(const ECDSA_SIG *a, char **pp);

        int DSA_do_verify(const char *dgst, int dgst_len, DSA_SIG *sig, DSA *dsa);
        int ECDSA_do_verify(const char *dgst, int dgst_len, const ECDSA_SIG *sig, EC_KEY *eckey);

        void DSA_SIG_free(DSA_SIG *a);
        void ECDSA_SIG_free(ECDSA_SIG *a);

        DSA *EVP_PKEY_get1_DSA(EVP_PKEY *pkey);
        EC_KEY *EVP_PKEY_get1_EC_KEY(EVP_PKEY *pkey);

        int RSA_verify_PKCS1_PSS(RSA *rsa, const char *mHash,
                        const EVP_MD *Hash, const char *EM,
                        int sLen);
        int RSA_padding_add_PKCS1_PSS(RSA *rsa, char *EM,
                        const char *mHash, const EVP_MD *Hash,
                        int sLen);

        int EVP_DigestInit_ex(EVP_MD_CTX *ctx, const EVP_MD *type, ENGINE *impl);
        int EVP_SignFinal(EVP_MD_CTX *ctx, char *sig, unsigned int *s, EVP_PKEY *pkey);
        int EVP_VerifyFinal(EVP_MD_CTX *ctx, char *sigbuf, unsigned int siglen, EVP_PKEY *pkey);

        void EVP_MD_CTX_set_flags(EVP_MD_CTX *ctx, int flags);
    """)
else:
    ffi.cdef("""
        int PKCS5_PBKDF2_HMAC(const char *pass, int passlen,
                        const char *salt, int saltlen, int iter,
                        const EVP_MD *digest,
                        int keylen, char *out);

        int EVP_DigestSignInit(EVP_MD_CTX *ctx, EVP_PKEY_CTX **pctx, const EVP_MD *type, ENGINE *e, EVP_PKEY *pkey);
        int EVP_DigestSignFinal(EVP_MD_CTX *ctx, char *sig, size_t *siglen);

        int EVP_DigestVerifyInit(EVP_MD_CTX *ctx, EVP_PKEY_CTX **pctx, const EVP_MD *type, ENGINE *e, EVP_PKEY *pkey);
        int EVP_DigestVerifyFinal(EVP_MD_CTX *ctx, const char *sig, size_t siglen);

        int EVP_PKEY_CTX_ctrl(EVP_PKEY_CTX *ctx, int keytype, int optype, int cmd, int p1, void *p2);
    """)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .. import ffi
from ._decode import _try_decode
from .._types import str_cls

if ffi() == 'cffi':
    from ._kernel32_cffi import kernel32, get_error
else:
    from ._kernel32_ctypes import kernel32, get_error


__all__ = [
    'handle_error',
    'kernel32',
]


def handle_error(result):
    """
    Extracts the last Windows error message into a python unicode string

    :param result:
        A function result, 0 or None indicates failure

    :return:
        A unicode string error message
    """

    if result:
        return

    _, error_string = get_error()

    if not isinstance(error_string, str_cls):
        error_string = _try_decode(error_string)

    raise OSError(error_string)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys

from .._ffi import register_ffi
from .._types import str_cls
from ..errors import LibraryNotFoundError

import cffi


__all__ = [
    'crypt32',
    'get_error',
]


ffi = cffi.FFI()
if cffi.__version_info__ >= (0, 9):
    ffi.set_unicode(True)
if sys.maxsize > 2 ** 32:
    ffi.cdef("typedef uint64_t ULONG_PTR;")
else:
    ffi.cdef("typedef unsigned long ULONG_PTR;")
ffi.cdef("""
    typedef HANDLE HCERTSTORE;
    typedef unsigned char *PBYTE;


    typedef struct _CRYPTOAPI_BLOB {
        DWORD cbData;
        PBYTE pbData;
    } CRYPTOAPI_BLOB;
    typedef CRYPTOAPI_BLOB CRYPT_INTEGER_BLOB;
    typedef CRYPTOAPI_BLOB CERT_NAME_BLOB;
    typedef CRYPTOAPI_BLOB CRYPT_BIT_BLOB;
    typedef CRYPTOAPI_BLOB CRYPT_OBJID_BLOB;

    typedef struct _CRYPT_ALGORITHM_IDENTIFIER {
        LPSTR pszObjId;
        CRYPT_OBJID_BLOB Parameters;
    } CRYPT_ALGORITHM_IDENTIFIER;

    typedef struct _FILETIME {
        DWORD dwLowDateTime;
        DWORD dwHighDateTime;
    } FILETIME;

    typedef struct _CERT_PUBLIC_KEY_INFO {
        CRYPT_ALGORITHM_IDENTIFIER Algorithm;
        CRYPT_BIT_BLOB PublicKey;
    } CERT_PUBLIC_KEY_INFO;

    typedef struct _CERT_EXTENSION {
        LPSTR pszObjId;
        BOOL fCritical;
        CRYPT_OBJID_BLOB Value;
    } CERT_EXTENSION, *PCERT_EXTENSION;

    typedef struct _CERT_INFO {
        DWORD dwVersion;
        CRYPT_INTEGER_BLOB SerialNumber;
        CRYPT_ALGORITHM_IDENTIFIER SignatureAlgorithm;
        CERT_NAME_BLOB Issuer;
        FILETIME NotBefore;
        FILETIME NotAfter;
        CERT_NAME_BLOB Subject;
        CERT_PUBLIC_KEY_INFO SubjectPublicKeyInfo;
        CRYPT_BIT_BLOB IssuerUniqueId;
        CRYPT_BIT_BLOB SubjectUniqueId;
        DWORD cExtension;
        PCERT_EXTENSION *rgExtension;
    } CERT_INFO, *PCERT_INFO;

    typedef struct _CERT_CONTEXT {
        DWORD dwCertEncodingType;
        PBYTE pbCertEncoded;
        DWORD cbCertEncoded;
        PCERT_INFO pCertInfo;
        HCERTSTORE hCertStore;
    } CERT_CONTEXT, *PCERT_CONTEXT;

    typedef struct _CERT_TRUST_STATUS {
        DWORD dwErrorStatus;
        DWORD dwInfoStatus;
    } CERT_TRUST_STATUS, *PCERT_TRUST_STATUS;

    typedef struct _CERT_ENHKEY_USAGE {
        DWORD cUsageIdentifier;
        LPSTR *rgpszUsageIdentifier;
    } CERT_ENHKEY_USAGE, *PCERT_ENHKEY_USAGE;

    typedef struct _CERT_CHAIN_ELEMENT {
        DWORD cbSize;
        PCERT_CONTEXT pCertContext;
        CERT_TRUST_STATUS TrustStatus;
        void *pRevocationInfo;
        PCERT_ENHKEY_USAGE pIssuanceUsage;
        PCERT_ENHKEY_USAGE pApplicationUsage;
        LPCWSTR pwszExtendedErrorInfo;
    } CERT_CHAIN_ELEMENT, *PCERT_CHAIN_ELEMENT;

    typedef struct _CERT_SIMPLE_CHAIN {
        DWORD cbSize;
        CERT_TRUST_STATUS TrustStatus;
        DWORD cElement;
        PCERT_CHAIN_ELEMENT *rgpElement;
        void *pTrustListInfo;
        BOOL fHasRevocationFreshnessTime;
        DWORD dwRevocationFreshnessTime;
    } CERT_SIMPLE_CHAIN, *PCERT_SIMPLE_CHAIN;

    typedef struct _CERT_CHAIN_CONTEXT {
        DWORD cbSize;
        CERT_TRUST_STATUS TrustStatus;
        DWORD cChain;
        PCERT_SIMPLE_CHAIN *rgpChain;
        DWORD cLowerQualityChainContext;
        void *rgpLowerQualityChainContext;
        BOOL fHasRevocationFreshnessTime;
        DWORD dwRevocationFreshnessTime;
    } CERT_CHAIN_CONTEXT, *PCERT_CHAIN_CONTEXT;

    typedef struct _CERT_USAGE_MATCH {
        DWORD dwType;
        CERT_ENHKEY_USAGE Usage;
    } CERT_USAGE_MATCH;

    typedef struct _CERT_CHAIN_PARA {
        DWORD cbSize;
        CERT_USAGE_MATCH RequestedUsage;
    } CERT_CHAIN_PARA;

    typedef struct _CERT_CHAIN_POLICY_PARA {
        DWORD cbSize;
        DWORD dwFlags;
        void  *pvExtraPolicyPara;
    } CERT_CHAIN_POLICY_PARA;

    typedef struct _HTTPSPolicyCallbackData {
        DWORD cbSize;
        DWORD dwAuthType;
        DWORD fdwChecks;
        WCHAR *pwszServerName;
    } SSL_EXTRA_CERT_CHAIN_POLICY_PARA;

    typedef struct _CERT_CHAIN_POLICY_STATUS {
        DWORD cbSize;
        DWORD dwError;
        LONG lChainIndex;
        LONG lElementIndex;
        void *pvExtraPolicyStatus;
    } CERT_CHAIN_POLICY_STATUS;

    typedef HANDLE HCERTCHAINENGINE;
    typedef HANDLE HCRYPTPROV;

    HCERTSTORE CertOpenStore(LPCSTR lpszStoreProvider, DWORD dwMsgAndCertEncodingType, HCRYPTPROV hCryptProv,
                    DWORD dwFlags, void *pvPara);
    BOOL CertAddEncodedCertificateToStore(HCERTSTORE hCertStore, DWORD dwCertEncodingType, BYTE *pbCertEncoded,
                    DWORD cbCertEncoded, DWORD dwAddDisposition, PCERT_CONTEXT *ppCertContext);
    BOOL CertGetCertificateChain(HCERTCHAINENGINE hChainEngine, CERT_CONTEXT *pCertContext, FILETIME *pTime,
                    HCERTSTORE hAdditionalStore, CERT_CHAIN_PARA *pChainPara, DWORD dwFlags, void *pvReserved,
                    PCERT_CHAIN_CONTEXT *ppChainContext);
    BOOL CertVerifyCertificateChainPolicy(ULONG_PTR pszPolicyOID, PCERT_CHAIN_CONTEXT pChainContext,
                    CERT_CHAIN_POLICY_PARA *pPolicyPara, CERT_CHAIN_POLICY_STATUS *pPolicyStatus);
    void CertFreeCertificateChain(PCERT_CHAIN_CONTEXT pChainContext);

    HCERTSTORE CertOpenSystemStoreW(HANDLE hprov, LPCWSTR szSubsystemProtocol);
    PCERT_CONTEXT CertEnumCertificatesInStore(HCERTSTORE hCertStore, CERT_CONTEXT *pPrevCertContext);
    BOOL CertCloseStore(HCERTSTORE hCertStore, DWORD dwFlags);
    BOOL CertGetEnhancedKeyUsage(CERT_CONTEXT *pCertContext, DWORD dwFlags, CERT_ENHKEY_USAGE *pUsage, DWORD *pcbUsage);
""")


try:
    crypt32 = ffi.dlopen('crypt32.dll')
    register_ffi(crypt32, ffi)

except (OSError) as e:
    if str_cls(e).find('cannot load library') != -1:
        raise LibraryNotFoundError('crypt32.dll could not be found')
    raise


def get_error():
    return ffi.getwinerror()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import platform
import sys
import threading

from ._types import str_cls, type_name
from .errors import LibraryNotFoundError
from .version import __version__, __version_info__


__all__ = [
    '__version__',
    '__version_info__',
    'backend',
    'ffi',
    'load_order',
    'use_ctypes',
    'use_openssl',
    'use_winlegacy',
]


_backend_lock = threading.Lock()
_module_values = {
    'backend': None,
    'backend_config': None,
    'ffi': None
}


def backend():
    """
    :return:
        A unicode string of the backend being used: "openssl", "mac", "win",
        "winlegacy"
    """

    if _module_values['backend'] is not None:
        return _module_values['backend']

    with _backend_lock:
        if _module_values['backend'] is not None:
            return _module_values['backend']

        if sys.platform == 'win32':
            # Windows XP was major version 5, Vista was 6
            if sys.getwindowsversion()[0] < 6:
                _module_values['backend'] = 'winlegacy'
            else:
                _module_values['backend'] = 'win'
        elif sys.platform == 'darwin':
            _module_values['backend'] = 'mac'
        else:
            _module_values['backend'] = 'openssl'

        return _module_values['backend']


def _backend_config():
    """
    :return:
        A dict of config info for the backend. Only currently used by "openssl",
        it may contains zero or more of the following keys:
         - "libcrypto_path"
         - "libssl_path"
    """

    if backend() != 'openssl':
        return {}

    if _module_values['backend_config'] is not None:
        return _module_values['backend_config']

    with _backend_lock:
        if _module_values['backend_config'] is not None:
            return _module_values['backend_config']

        _module_values['backend_config'] = {}
        return _module_values['backend_config']


def use_openssl(libcrypto_path, libssl_path, trust_list_path=None):
    """
    Forces using OpenSSL dynamic libraries on OS X (.dylib) or Windows (.dll),
    or using a specific dynamic library on Linux/BSD (.so).

    This can also be used to configure oscrypto to use LibreSSL dynamic
    libraries.

    This method must be called before any oscrypto submodules are imported.

    :param libcrypto_path:
        A unicode string of the file path to the OpenSSL/LibreSSL libcrypto
        dynamic library.

    :param libssl_path:
        A unicode string of the file path to the OpenSSL/LibreSSL libssl
        dynamic library.

    :param trust_list_path:
        An optional unicode string of the path to a file containing
        OpenSSL-compatible CA certificates in PEM format. If this is not
        provided and the platform is OS X or Windows, the system trust roots
        will be exported from the OS and used for all TLS connections.

    :raises:
        ValueError - when one of the paths is not a unicode string
        OSError - when the trust_list_path does not exist on the filesystem
        oscrypto.errors.LibraryNotFoundError - when one of the path does not exist on the filesystem
        RuntimeError - when this function is called after another part of oscrypto has been imported
    """

    if not isinstance(libcrypto_path, str_cls):
        raise ValueError('libcrypto_path must be a unicode string, not %s' % type_name(libcrypto_path))

    if not isinstance(libssl_path, str_cls):
        raise ValueError('libssl_path must be a unicode string, not %s' % type_name(libssl_path))

    if not os.path.exists(libcrypto_path):
        raise LibraryNotFoundError('libcrypto does not exist at %s' % libcrypto_path)

    if not os.path.exists(libssl_path):
        raise LibraryNotFoundError('libssl does not exist at %s' % libssl_path)

    if trust_list_path is not None:
        if not isinstance(trust_list_path, str_cls):
            raise ValueError('trust_list_path must be a unicode string, not %s' % type_name(trust_list_path))
        if not os.path.exists(trust_list_path):
            raise OSError('trust_list_path does not exist at %s' % trust_list_path)

    with _backend_lock:
        new_config = {
            'libcrypto_path': libcrypto_path,
            'libssl_path': libssl_path,
            'trust_list_path': trust_list_path,
        }

        if _module_values['backend'] == 'openssl' and _module_values['backend_config'] == new_config:
            return

        if _module_values['backend'] is not None:
            raise RuntimeError('Another part of oscrypto has already been imported, unable to force use of OpenSSL')

        _module_values['backend'] = 'openssl'
        _module_values['backend_config'] = new_config


def use_winlegacy():
    """
    Forces use of the legacy Windows CryptoAPI. This should only be used on
    Windows XP or for testing. It is less full-featured than the Cryptography
    Next Generation (CNG) API, and as a result the elliptic curve and PSS
    padding features are implemented in pure Python. This isn't ideal, but it
    a shim for end-user client code. No one is going to run a server on Windows
    XP anyway, right?!

    :raises:
        EnvironmentError - when this function is called on an operating system other than Windows
        RuntimeError - when this function is called after another part of oscrypto has been imported
    """

    if sys.platform != 'win32':
        plat = platform.system() or sys.platform
        if plat == 'Darwin':
            plat = 'OS X'
        raise EnvironmentError('The winlegacy backend can only be used on Windows, not %s' % plat)

    with _backend_lock:
        if _module_values['backend'] == 'winlegacy':
            return

        if _module_values['backend'] is not None:
            raise RuntimeError(
                'Another part of oscrypto has already been imported, unable to force use of Windows legacy CryptoAPI'
            )

        _module_values['backend'] = 'winlegacy'


def use_ctypes():
    """
    Forces use of ctypes instead of cffi for the FFI layer

    :raises:
        RuntimeError - when this function is called after another part of oscrypto has been imported
    """

    with _backend_lock:
        if _module_values['ffi'] == 'ctypes':
            return

        if _module_values['backend'] is not None:
            raise RuntimeError(
                'Another part of oscrypto has already been imported, unable to force use of ctypes'
            )

        _module_values['ffi'] = 'ctypes'


def ffi():
    """
    Returns the FFI module being used

    :return:
        A unicode string of "cffi" or "ctypes"
    """

    if _module_values['ffi'] is not None:
        return _module_values['ffi']

    with _backend_lock:
        try:
            import cffi  # noqa: F401
            _module_values['ffi'] = 'cffi'
        except (ImportError):
            _module_values['ffi'] = 'ctypes'

        return _module_values['ffi']


def load_order():
    """
    Returns a list of the module and sub-module names for oscrypto in
    dependency load order, for the sake of live reloading code

    :return:
        A list of unicode strings of module names, as they would appear in
        sys.modules, ordered by which module should be reloaded first
    """

    return [
        'oscrypto._asn1',
        'oscrypto._cipher_suites',
        'oscrypto._errors',
        'oscrypto._int',
        'oscrypto._types',
        'oscrypto.errors',
        'oscrypto.version',
        'oscrypto',
        'oscrypto._ffi',
        'oscrypto._pkcs12',
        'oscrypto._pkcs5',
        'oscrypto._rand',
        'oscrypto._tls',
        'oscrypto._linux_bsd.trust_list',
        'oscrypto._mac._common_crypto_cffi',
        'oscrypto._mac._common_crypto_ctypes',
        'oscrypto._mac._common_crypto',
        'oscrypto._mac._core_foundation_cffi',
        'oscrypto._mac._core_foundation_ctypes',
        'oscrypto._mac._core_foundation',
        'oscrypto._mac._security_cffi',
        'oscrypto._mac._security_ctypes',
        'oscrypto._mac._security',
        'oscrypto._mac.trust_list',
        'oscrypto._mac.util',
        'oscrypto._openssl._libcrypto_cffi',
        'oscrypto._openssl._libcrypto_ctypes',
        'oscrypto._openssl._libcrypto',
        'oscrypto._openssl._libssl_cffi',
        'oscrypto._openssl._libssl_ctypes',
        'oscrypto._openssl._libssl',
        'oscrypto._openssl.util',
        'oscrypto._win._cng_cffi',
        'oscrypto._win._cng_ctypes',
        'oscrypto._win._cng',
        'oscrypto._win._decode',
        'oscrypto._win._advapi32_cffi',
        'oscrypto._win._advapi32_ctypes',
        'oscrypto._win._advapi32',
        'oscrypto._win._kernel32_cffi',
        'oscrypto._win._kernel32_ctypes',
        'oscrypto._win._kernel32',
        'oscrypto._win._secur32_cffi',
        'oscrypto._win._secur32_ctypes',
        'oscrypto._win._secur32',
        'oscrypto._win._crypt32_cffi',
        'oscrypto._win._crypt32_ctypes',
        'oscrypto._win._crypt32',
        'oscrypto._win.trust_list',
        'oscrypto._win.util',
        'oscrypto.trust_list',
        'oscrypto.util',
        'oscrypto.kdf',
        'oscrypto._mac.symmetric',
        'oscrypto._openssl.symmetric',
        'oscrypto._win.symmetric',
        'oscrypto.symmetric',
        'oscrypto._asymmetric',
        'oscrypto._ecdsa',
        'oscrypto._pkcs1',
        'oscrypto._mac.asymmetric',
        'oscrypto._openssl.asymmetric',
        'oscrypto._win.asymmetric',
        'oscrypto.asymmetric',
        'oscrypto.keys',
        'oscrypto._mac.tls',
        'oscrypto._openssl.tls',
        'oscrypto._win.tls',
        'oscrypto.tls',
    ]



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import hashlib
import hmac
import sys

from . import backend
from ._asn1 import (
    Certificate,
    DSASignature,
    ECDomainParameters,
    ECPointBitString,
    ECPrivateKey,
    int_from_bytes,
    PrivateKeyAlgorithm,
    PrivateKeyInfo,
    PublicKeyAlgorithm,
    PublicKeyInfo,
)
from ._errors import pretty_message
from ._types import type_name, byte_cls
from .util import rand_bytes
from .errors import SignatureError

if sys.version_info < (3,):
    chr_cls = chr
    range = xrange  # noqa

else:
    def chr_cls(num):
        return bytes([num])


_backend = backend()


if _backend != 'winlegacy':
    # This pure-Python ECDSA code is only suitable for use on client machines,
    # and is only needed on Windows 5.x (XP/2003). For testing sake it is
    # possible to force use of it on newer versions of Windows.
    raise SystemError('Pure-python ECDSA code is only for Windows XP/2003')


__all__ = [
    'ec_generate_pair',
    'ec_compute_public_key_point',
    'ec_public_key_info',
    'ecdsa_sign',
    'ecdsa_verify',
]


CURVE_BYTES = {
    'secp256r1': 32,
    'secp384r1': 48,
    'secp521r1': 66,
}

CURVE_EXTRA_BITS = {
    'secp256r1': 0,
    'secp384r1': 0,
    'secp521r1': 7,
}


def ec_generate_pair(curve):
    """
    Generates a EC public/private key pair

    :param curve:
        A unicode string. Valid values include "secp256r1", "secp384r1" and
        "secp521r1".

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type

    :return:
        A 2-element tuple of (asn1crypto.keys.PublicKeyInfo,
        asn1crypto.keys.PrivateKeyInfo)
    """

    if curve not in set(['secp256r1', 'secp384r1', 'secp521r1']):
        raise ValueError(pretty_message(
            '''
            curve must be one of "secp256r1", "secp384r1", "secp521r1", not %s
            ''',
            repr(curve)
        ))

    curve_num_bytes = CURVE_BYTES[curve]
    curve_base_point = {
        'secp256r1': SECP256R1_BASE_POINT,
        'secp384r1': SECP384R1_BASE_POINT,
        'secp521r1': SECP521R1_BASE_POINT,
    }[curve]

    while True:
        private_key_bytes = rand_bytes(curve_num_bytes)
        private_key_int = int_from_bytes(private_key_bytes, signed=False)

        if private_key_int > 0 and private_key_int < curve_base_point.order:
            break

    private_key_info = PrivateKeyInfo({
        'version': 0,
        'private_key_algorithm': PrivateKeyAlgorithm({
            'algorithm': 'ec',
            'parameters': ECDomainParameters(
                name='named',
                value=curve
            )
        }),
        'private_key': ECPrivateKey({
            'version': 'ecPrivkeyVer1',
            'private_key': private_key_int
        }),
    })

    ec_point = ec_compute_public_key_point(private_key_info)
    private_key_info['private_key'].parsed['public_key'] = ec_point.copy()

    return (ec_public_key_info(ec_point, curve), private_key_info)


def ec_compute_public_key_point(private_key):
    """
    Constructs the PublicKeyInfo for a PrivateKeyInfo

    :param private_key:
        An asn1crypto.keys.PrivateKeyInfo object

    :raises:
        ValueError - when any of the parameters contain an invalid value

    :return:
        An asn1crypto.keys.ECPointBitString object
    """

    if not isinstance(private_key, PrivateKeyInfo):
        raise TypeError(pretty_message(
            '''
            private_key must be an instance of the
            asn1crypto.keys.PrivateKeyInfo class, not %s
            ''',
            type_name(private_key)
        ))

    curve_type, details = private_key.curve

    if curve_type == 'implicit_ca':
        raise ValueError(pretty_message(
            '''
            Unable to compute public key for EC key using Implicit CA
            parameters
            '''
        ))

    if curve_type == 'specified':
        raise ValueError(pretty_message(
            '''
            Unable to compute public key for EC key over a specified field
            '''
        ))

    elif curve_type == 'named':
        if details not in set(['secp256r1', 'secp384r1', 'secp521r1']):
            raise ValueError(pretty_message(
                '''
                Named curve must be one of "secp256r1", "secp384r1", "secp521r1", not %s
                ''',
                repr(details)
            ))

        base_point = {
            'secp256r1': SECP256R1_BASE_POINT,
            'secp384r1': SECP384R1_BASE_POINT,
            'secp521r1': SECP521R1_BASE_POINT,
        }[details]

    public_point = base_point * private_key['private_key'].parsed['private_key'].native
    return ECPointBitString.from_coords(public_point.x, public_point.y)


def ec_public_key_info(public_key_point, curve):
    """
    Constructs the PublicKeyInfo for an ECPointBitString

    :param private_key:
        An asn1crypto.keys.ECPointBitString object

    :param curve:
        A unicode string of the curve name - one of secp256r1, secp384r1 or secp521r1

    :raises:
        ValueError - when any of the parameters contain an invalid value

    :return:
        An asn1crypto.keys.PublicKeyInfo object
    """

    if curve not in set(['secp256r1', 'secp384r1', 'secp521r1']):
        raise ValueError(pretty_message(
            '''
            curve must be one of "secp256r1", "secp384r1", "secp521r1", not %s
            ''',
            repr(curve)
        ))

    return PublicKeyInfo({
        'algorithm': PublicKeyAlgorithm({
            'algorithm': 'ec',
            'parameters': ECDomainParameters(
                name='named',
                value=curve
            )
        }),
        'public_key': public_key_point,
    })


def ecdsa_sign(private_key, data, hash_algorithm):
    """
    Generates an ECDSA signature in pure Python (thus slow)

    :param private_key:
        The PrivateKey to generate the signature with

    :param data:
        A byte string of the data the signature is for

    :param hash_algorithm:
        A unicode string of "sha1", "sha256", "sha384" or "sha512"

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the signature
    """

    if not hasattr(private_key, 'asn1') or not isinstance(private_key.asn1, PrivateKeyInfo):
        raise TypeError(pretty_message(
            '''
            private_key must be an instance of the
            oscrypto.asymmetric.PrivateKey class, not %s
            ''',
            type_name(private_key)
        ))

    curve_name = private_key.curve
    if curve_name not in set(['secp256r1', 'secp384r1', 'secp521r1']):
        raise ValueError(pretty_message(
            '''
            private_key does not use one of the named curves secp256r1,
            secp384r1 or secp521r1
            '''
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if hash_algorithm not in set(['sha1', 'sha224', 'sha256', 'sha384', 'sha512']):
        raise ValueError(pretty_message(
            '''
            hash_algorithm must be one of "sha1", "sha224", "sha256", "sha384",
            "sha512", not %s
            ''',
            repr(hash_algorithm)
        ))

    hash_func = getattr(hashlib, hash_algorithm)

    ec_private_key = private_key.asn1['private_key'].parsed
    private_key_bytes = ec_private_key['private_key'].contents
    private_key_int = ec_private_key['private_key'].native

    curve_num_bytes = CURVE_BYTES[curve_name]
    curve_base_point = {
        'secp256r1': SECP256R1_BASE_POINT,
        'secp384r1': SECP384R1_BASE_POINT,
        'secp521r1': SECP521R1_BASE_POINT,
    }[curve_name]

    n = curve_base_point.order

    # RFC 6979 section 3.2

    # a.
    digest = hash_func(data).digest()
    hash_length = len(digest)

    h = int_from_bytes(digest, signed=False) % n

    # b.
    V = b'\x01' * hash_length

    # c.
    K = b'\x00' * hash_length

    # d.
    K = hmac.new(K, V + b'\x00' + private_key_bytes + digest, hash_func).digest()

    # e.
    V = hmac.new(K, V, hash_func).digest()

    # f.
    K = hmac.new(K, V + b'\x01' + private_key_bytes + digest, hash_func).digest()

    # g.
    V = hmac.new(K, V, hash_func).digest()

    # h.
    r = 0
    s = 0
    while True:
        # h. 1
        T = b''

        # h. 2
        while len(T) < curve_num_bytes:
            V = hmac.new(K, V, hash_func).digest()
            T += V

        # h. 3
        k = int_from_bytes(T[0:curve_num_bytes], signed=False)
        if k == 0 or k >= n:
            continue

        # Calculate the signature in the loop in case we need a new k
        r = (curve_base_point * k).x % n
        if r == 0:
            continue

        s = (inverse_mod(k, n) * (h + (private_key_int * r) % n)) % n
        if s == 0:
            continue

        break

    return DSASignature({'r': r, 's': s}).dump()


def ecdsa_verify(certificate_or_public_key, signature, data, hash_algorithm):
    """
    Verifies an ECDSA signature in pure Python (thus slow)

    :param certificate_or_public_key:
        A Certificate or PublicKey instance to verify the signature with

    :param signature:
        A byte string of the signature to verify

    :param data:
        A byte string of the data the signature is for

    :param hash_algorithm:
        A unicode string of "md5", "sha1", "sha256", "sha384" or "sha512"

    :raises:
        oscrypto.errors.SignatureError - when the signature is determined to be invalid
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library
    """

    has_asn1 = hasattr(certificate_or_public_key, 'asn1')
    if not has_asn1 or not isinstance(certificate_or_public_key.asn1, (PublicKeyInfo, Certificate)):
        raise TypeError(pretty_message(
            '''
            certificate_or_public_key must be an instance of the
            oscrypto.asymmetric.PublicKey or oscrypto.asymmetric.Certificate
            classes, not %s
            ''',
            type_name(certificate_or_public_key)
        ))

    curve_name = certificate_or_public_key.curve
    if curve_name not in set(['secp256r1', 'secp384r1', 'secp521r1']):
        raise ValueError(pretty_message(
            '''
            certificate_or_public_key does not use one of the named curves
            secp256r1, secp384r1 or secp521r1
            '''
        ))

    if not isinstance(signature, byte_cls):
        raise TypeError(pretty_message(
            '''
            signature must be a byte string, not %s
            ''',
            type_name(signature)
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if hash_algorithm not in set(['sha1', 'sha224', 'sha256', 'sha384', 'sha512']):
        raise ValueError(pretty_message(
            '''
            hash_algorithm must be one of "sha1", "sha224", "sha256", "sha384",
            "sha512", not %s
            ''',
            repr(hash_algorithm)
        ))

    asn1 = certificate_or_public_key.asn1
    if isinstance(asn1, Certificate):
        asn1 = asn1.public_key

    curve_base_point = {
        'secp256r1': SECP256R1_BASE_POINT,
        'secp384r1': SECP384R1_BASE_POINT,
        'secp521r1': SECP521R1_BASE_POINT,
    }[curve_name]

    x, y = asn1['public_key'].to_coords()
    n = curve_base_point.order

    # Validates that the point is valid
    public_key_point = PrimePoint(curve_base_point.curve, x, y, n)

    try:
        signature = DSASignature.load(signature)
        r = signature['r'].native
        s = signature['s'].native
    except (ValueError):
        raise SignatureError('Signature is invalid')

    invalid = 0

    # Check r is valid
    invalid |= r < 1
    invalid |= r >= n

    # Check s is valid
    invalid |= s < 1
    invalid |= s >= n

    if invalid:
        raise SignatureError('Signature is invalid')

    hash_func = getattr(hashlib, hash_algorithm)

    digest = hash_func(data).digest()

    z = int_from_bytes(digest, signed=False) % n
    w = inverse_mod(s, n)
    u1 = (z * w) % n
    u2 = (r * w) % n
    hash_point = (curve_base_point * u1) + (public_key_point * u2)
    if r != (hash_point.x % n):
        raise SignatureError('Signature is invalid')


"""
Classes and objects to represent prime-field elliptic curves and points on them.
Exports the following items:

 - PrimeCurve()
 - PrimePoint()
 - SECP192R1_CURVE
 - SECP192R1_BASE_POINT
 - SECP224R1_CURVE
 - SECP224R1_BASE_POINT
 - SECP256R1_CURVE
 - SECP256R1_BASE_POINT
 - SECP384R1_CURVE
 - SECP384R1_BASE_POINT
 - SECP521R1_CURVE
 - SECP521R1_BASE_POINT

The curve constants are all PrimeCurve() objects and the base point constants
are all PrimePoint() objects.

Some of the following source code is derived from
http://webpages.charter.net/curryfans/peter/downloads.html, but has been heavily
modified to fit into this projects lint settings. The original project license
is listed below:

Copyright (c) 2014 Peter Pearson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


def inverse_mod(a, p):
    """
    Compute the modular inverse of a (mod p)

    :param a:
        An integer

    :param p:
        An integer

    :return:
        An integer
    """

    if a < 0 or p <= a:
        a = a % p

    # From Ferguson and Schneier, roughly:

    c, d = a, p
    uc, vc, ud, vd = 1, 0, 0, 1
    while c != 0:
        q, c, d = divmod(d, c) + (c,)
        uc, vc, ud, vd = ud - q * uc, vd - q * vc, uc, vc

    # At this point, d is the GCD, and ud*a+vd*p = d.
    # If d == 1, this means that ud is a inverse.

    assert d == 1
    if ud > 0:
        return ud
    else:
        return ud + p


class PrimeCurve():
    """
    Elliptic curve over a prime field. Characteristic two field curves are not
    supported.
    """

    def __init__(self, p, a, b):
        """
        The curve of points satisfying y^2 = x^3 + a*x + b (mod p)

        :param p:
            The prime number as an integer

        :param a:
            The component a as an integer

        :param b:
            The component b as an integer
        """

        self.p = p
        self.a = a
        self.b = b

    def contains(self, point):
        """
        :param point:
            A Point object

        :return:
            Boolean if the point is on this curve
        """

        y2 = point.y * point.y
        x3 = point.x * point.x * point.x
        return (y2 - (x3 + self.a * point.x + self.b)) % self.p == 0


class PrimePoint():
    """
    A point on a prime-field elliptic curve
    """

    def __init__(self, curve, x, y, order=None):
        """
        :param curve:
            A PrimeCurve object

        :param x:
            The x coordinate of the point as an integer

        :param y:
            The y coordinate of the point as an integer

        :param order:
            The order of the point, as an integer - optional
        """

        self.curve = curve
        self.x = x
        self.y = y
        self.order = order

        # self.curve is allowed to be None only for INFINITY:
        if self.curve:
            if not self.curve.contains(self):
                raise ValueError('Invalid EC point')

        if self.order:
            if self * self.order != INFINITY:
                raise ValueError('Invalid EC point')

    def __cmp__(self, other):
        """
        :param other:
            A PrimePoint object

        :return:
            0 if identical, 1 otherwise
        """
        if self.curve == other.curve and self.x == other.x and self.y == other.y:
            return 0
        else:
            return 1

    def __add__(self, other):
        """
        :param other:
            A PrimePoint object

        :return:
            A PrimePoint object
        """

        # X9.62 B.3:

        if other == INFINITY:
            return self
        if self == INFINITY:
            return other
        assert self.curve == other.curve
        if self.x == other.x:
            if (self.y + other.y) % self.curve.p == 0:
                return INFINITY
            else:
                return self.double()

        p = self.curve.p

        l_ = ((other.y - self.y) * inverse_mod(other.x - self.x, p)) % p

        x3 = (l_ * l_ - self.x - other.x) % p
        y3 = (l_ * (self.x - x3) - self.y) % p

        return PrimePoint(self.curve, x3, y3)

    def __mul__(self, other):
        """
        :param other:
            An integer to multiple the Point by

        :return:
            A PrimePoint object
        """

        def leftmost_bit(x):
            assert x > 0
            result = 1
            while result <= x:
                result = 2 * result
            return result // 2

        e = other
        if self.order:
            e = e % self.order
        if e == 0:
            return INFINITY
        if self == INFINITY:
            return INFINITY
        assert e > 0

        # From X9.62 D.3.2:

        e3 = 3 * e
        negative_self = PrimePoint(self.curve, self.x, -self.y, self.order)
        i = leftmost_bit(e3) // 2
        result = self
        # print "Multiplying %s by %d (e3 = %d):" % ( self, other, e3 )
        while i > 1:
            result = result.double()
            if (e3 & i) != 0 and (e & i) == 0:
                result = result + self
            if (e3 & i) == 0 and (e & i) != 0:
                result = result + negative_self
            # print ". . . i = %d, result = %s" % ( i, result )
            i = i // 2

        return result

    def __rmul__(self, other):
        """
        :param other:
            An integer to multiple the Point by

        :return:
            A PrimePoint object
        """

        return self * other

    def double(self):
        """
        :return:
            A PrimePoint object that is twice this point
        """

        # X9.62 B.3:

        p = self.curve.p
        a = self.curve.a

        l_ = ((3 * self.x * self.x + a) * inverse_mod(2 * self.y, p)) % p

        x3 = (l_ * l_ - 2 * self.x) % p
        y3 = (l_ * (self.x - x3) - self.y) % p

        return PrimePoint(self.curve, x3, y3)


# This one point is the Point At Infinity for all purposes:
INFINITY = PrimePoint(None, None, None)


# NIST Curve P-192:
SECP192R1_CURVE = PrimeCurve(
    6277101735386680763835789423207666416083908700390324961279,
    -3,
    0x64210519e59c80e70fa7e9ab72243049feb8deecc146b9b1
)
SECP192R1_BASE_POINT = PrimePoint(
    SECP192R1_CURVE,
    0x188da80eb03090f67cbf20eb43a18800f4ff0afd82ff1012,
    0x07192b95ffc8da78631011ed6b24cdd573f977a11e794811,
    6277101735386680763835789423176059013767194773182842284081
)


# NIST Curve P-224:
SECP224R1_CURVE = PrimeCurve(
    26959946667150639794667015087019630673557916260026308143510066298881,
    -3,
    0xb4050a850c04b3abf54132565044b0b7d7bfd8ba270b39432355ffb4
)
SECP224R1_BASE_POINT = PrimePoint(
    SECP224R1_CURVE,
    0xb70e0cbd6bb4bf7f321390b94a03c1d356c21122343280d6115c1d21,
    0xbd376388b5f723fb4c22dfe6cd4375a05a07476444d5819985007e34,
    26959946667150639794667015087019625940457807714424391721682722368061
)


# NIST Curve P-256:
SECP256R1_CURVE = PrimeCurve(
    115792089210356248762697446949407573530086143415290314195533631308867097853951,
    -3,
    0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b
)
SECP256R1_BASE_POINT = PrimePoint(
    SECP256R1_CURVE,
    0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296,
    0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5,
    115792089210356248762697446949407573529996955224135760342422259061068512044369
)


# NIST Curve P-384:
SECP384R1_CURVE = PrimeCurve(
    39402006196394479212279040100143613805079739270465446667948293404245721771496870329047266088258938001861606973112319,  # noqa
    -3,
    0xb3312fa7e23ee7e4988e056be3f82d19181d9c6efe8141120314088f5013875ac656398d8a2ed19d2a85c8edd3ec2aef
)
SECP384R1_BASE_POINT = PrimePoint(
    SECP384R1_CURVE,
    0xaa87ca22be8b05378eb1c71ef320ad746e1d3b628ba79b9859f741e082542a385502f25dbf55296c3a545e3872760ab7,
    0x3617de4a96262c6f5d9e98bf9292dc29f8f41dbd289a147ce9da3113b5f0b8c00a60b1ce1d7e819d7a431d7c90ea0e5f,
    39402006196394479212279040100143613805079739270465446667946905279627659399113263569398956308152294913554433653942643
)


# NIST Curve P-521:
SECP521R1_CURVE = PrimeCurve(
    6864797660130609714981900799081393217269435300143305409394463459185543183397656052122559640661454554977296311391480858037121987999716643812574028291115057151,  # noqa
    -3,
    0x051953eb9618e1c9a1f929a21a0b68540eea2da725b99b315f3b8b489918ef109e156193951ec7e937b1652c0bd3bb1bf073573df883d2c34f1ef451fd46b503f00  # noqa
)
SECP521R1_BASE_POINT = PrimePoint(
    SECP521R1_CURVE,
    0xc6858e06b70404e9cd9e3ecb662395b4429c648139053fb521f828af606b4d3dbaa14b5e77efe75928fe1dc127a2ffa8de3348b3c1856a429bf97e7e31c2e5bd66,  # noqa
    0x11839296a789a3bc0045c8a5fb42c7d1bd998f54449579b446817afbd17273e662c97ee72995ef42640c550b9013fad0761353c7086a272c24088be94769fd16650,  # noqa
    6864797660130609714981900799081393217269435300143305409394463459185543183397655394245057746333217197532963996371363321113864768612440380340372808892707005449  # noqa
)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function


__all__ = [
    'CIPHER_SUITE_MAP',
]


CIPHER_SUITE_MAP = {
    b'\x00\x00': 'TLS_NULL_WITH_NULL_NULL',
    b'\x00\x01': 'TLS_RSA_WITH_NULL_MD5',
    b'\x00\x02': 'TLS_RSA_WITH_NULL_SHA',
    b'\x00\x03': 'TLS_RSA_EXPORT_WITH_RC4_40_MD5',
    b'\x00\x04': 'TLS_RSA_WITH_RC4_128_MD5',
    b'\x00\x05': 'TLS_RSA_WITH_RC4_128_SHA',
    b'\x00\x06': 'TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5',
    b'\x00\x07': 'TLS_RSA_WITH_IDEA_CBC_SHA',
    b'\x00\x08': 'TLS_RSA_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x09': 'TLS_RSA_WITH_DES_CBC_SHA',
    b'\x00\x0A': 'TLS_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x0B': 'TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x0C': 'TLS_DH_DSS_WITH_DES_CBC_SHA',
    b'\x00\x0D': 'TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x0E': 'TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x0F': 'TLS_DH_RSA_WITH_DES_CBC_SHA',
    b'\x00\x10': 'TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x11': 'TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x12': 'TLS_DHE_DSS_WITH_DES_CBC_SHA',
    b'\x00\x13': 'TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x14': 'TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x15': 'TLS_DHE_RSA_WITH_DES_CBC_SHA',
    b'\x00\x16': 'TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x17': 'TLS_DH_anon_EXPORT_WITH_RC4_40_MD5',
    b'\x00\x18': 'TLS_DH_anon_WITH_RC4_128_MD5',
    b'\x00\x19': 'TLS_DH_anon_EXPORT_WITH_DES40_CBC_SHA',
    b'\x00\x1A': 'TLS_DH_anon_WITH_DES_CBC_SHA',
    b'\x00\x1B': 'TLS_DH_anon_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x1E': 'TLS_KRB5_WITH_DES_CBC_SHA',
    b'\x00\x1F': 'TLS_KRB5_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x20': 'TLS_KRB5_WITH_RC4_128_SHA',
    b'\x00\x21': 'TLS_KRB5_WITH_IDEA_CBC_SHA',
    b'\x00\x22': 'TLS_KRB5_WITH_DES_CBC_MD5',
    b'\x00\x23': 'TLS_KRB5_WITH_3DES_EDE_CBC_MD5',
    b'\x00\x24': 'TLS_KRB5_WITH_RC4_128_MD5',
    b'\x00\x25': 'TLS_KRB5_WITH_IDEA_CBC_MD5',
    b'\x00\x26': 'TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA',
    b'\x00\x27': 'TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA',
    b'\x00\x28': 'TLS_KRB5_EXPORT_WITH_RC4_40_SHA',
    b'\x00\x29': 'TLS_KRB5_EXPORT_WITH_DES_CBC_40_MD5',
    b'\x00\x2A': 'TLS_KRB5_EXPORT_WITH_RC2_CBC_40_MD5',
    b'\x00\x2B': 'TLS_KRB5_EXPORT_WITH_RC4_40_MD5',
    b'\x00\x2C': 'TLS_PSK_WITH_NULL_SHA',
    b'\x00\x2D': 'TLS_DHE_PSK_WITH_NULL_SHA',
    b'\x00\x2E': 'TLS_RSA_PSK_WITH_NULL_SHA',
    b'\x00\x2F': 'TLS_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x30': 'TLS_DH_DSS_WITH_AES_128_CBC_SHA',
    b'\x00\x31': 'TLS_DH_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x32': 'TLS_DHE_DSS_WITH_AES_128_CBC_SHA',
    b'\x00\x33': 'TLS_DHE_RSA_WITH_AES_128_CBC_SHA',
    b'\x00\x34': 'TLS_DH_anon_WITH_AES_128_CBC_SHA',
    b'\x00\x35': 'TLS_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x36': 'TLS_DH_DSS_WITH_AES_256_CBC_SHA',
    b'\x00\x37': 'TLS_DH_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x38': 'TLS_DHE_DSS_WITH_AES_256_CBC_SHA',
    b'\x00\x39': 'TLS_DHE_RSA_WITH_AES_256_CBC_SHA',
    b'\x00\x3A': 'TLS_DH_anon_WITH_AES_256_CBC_SHA',
    b'\x00\x3B': 'TLS_RSA_WITH_NULL_SHA256',
    b'\x00\x3C': 'TLS_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x3D': 'TLS_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x3E': 'TLS_DH_DSS_WITH_AES_128_CBC_SHA256',
    b'\x00\x3F': 'TLS_DH_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x40': 'TLS_DHE_DSS_WITH_AES_128_CBC_SHA256',
    b'\x00\x41': 'TLS_RSA_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x42': 'TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x43': 'TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x44': 'TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x45': 'TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x46': 'TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA',
    b'\x00\x67': 'TLS_DHE_RSA_WITH_AES_128_CBC_SHA256',
    b'\x00\x68': 'TLS_DH_DSS_WITH_AES_256_CBC_SHA256',
    b'\x00\x69': 'TLS_DH_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x6A': 'TLS_DHE_DSS_WITH_AES_256_CBC_SHA256',
    b'\x00\x6B': 'TLS_DHE_RSA_WITH_AES_256_CBC_SHA256',
    b'\x00\x6C': 'TLS_DH_anon_WITH_AES_128_CBC_SHA256',
    b'\x00\x6D': 'TLS_DH_anon_WITH_AES_256_CBC_SHA256',
    b'\x00\x84': 'TLS_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x85': 'TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x86': 'TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x87': 'TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x88': 'TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x89': 'TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA',
    b'\x00\x8A': 'TLS_PSK_WITH_RC4_128_SHA',
    b'\x00\x8B': 'TLS_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x8C': 'TLS_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x8D': 'TLS_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x8E': 'TLS_DHE_PSK_WITH_RC4_128_SHA',
    b'\x00\x8F': 'TLS_DHE_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x90': 'TLS_DHE_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x91': 'TLS_DHE_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x92': 'TLS_RSA_PSK_WITH_RC4_128_SHA',
    b'\x00\x93': 'TLS_RSA_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\x00\x94': 'TLS_RSA_PSK_WITH_AES_128_CBC_SHA',
    b'\x00\x95': 'TLS_RSA_PSK_WITH_AES_256_CBC_SHA',
    b'\x00\x96': 'TLS_RSA_WITH_SEED_CBC_SHA',
    b'\x00\x97': 'TLS_DH_DSS_WITH_SEED_CBC_SHA',
    b'\x00\x98': 'TLS_DH_RSA_WITH_SEED_CBC_SHA',
    b'\x00\x99': 'TLS_DHE_DSS_WITH_SEED_CBC_SHA',
    b'\x00\x9A': 'TLS_DHE_RSA_WITH_SEED_CBC_SHA',
    b'\x00\x9B': 'TLS_DH_anon_WITH_SEED_CBC_SHA',
    b'\x00\x9C': 'TLS_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\x9D': 'TLS_RSA_WITH_AES_256_GCM_SHA384',
    b'\x00\x9E': 'TLS_DHE_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\x9F': 'TLS_DHE_RSA_WITH_AES_256_GCM_SHA384',
    b'\x00\xA0': 'TLS_DH_RSA_WITH_AES_128_GCM_SHA256',
    b'\x00\xA1': 'TLS_DH_RSA_WITH_AES_256_GCM_SHA384',
    b'\x00\xA2': 'TLS_DHE_DSS_WITH_AES_128_GCM_SHA256',
    b'\x00\xA3': 'TLS_DHE_DSS_WITH_AES_256_GCM_SHA384',
    b'\x00\xA4': 'TLS_DH_DSS_WITH_AES_128_GCM_SHA256',
    b'\x00\xA5': 'TLS_DH_DSS_WITH_AES_256_GCM_SHA384',
    b'\x00\xA6': 'TLS_DH_anon_WITH_AES_128_GCM_SHA256',
    b'\x00\xA7': 'TLS_DH_anon_WITH_AES_256_GCM_SHA384',
    b'\x00\xA8': 'TLS_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xA9': 'TLS_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\xAA': 'TLS_DHE_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xAB': 'TLS_DHE_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\xAC': 'TLS_RSA_PSK_WITH_AES_128_GCM_SHA256',
    b'\x00\xAD': 'TLS_RSA_PSK_WITH_AES_256_GCM_SHA384',
    b'\x00\xAE': 'TLS_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xAF': 'TLS_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xB0': 'TLS_PSK_WITH_NULL_SHA256',
    b'\x00\xB1': 'TLS_PSK_WITH_NULL_SHA384',
    b'\x00\xB2': 'TLS_DHE_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xB3': 'TLS_DHE_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xB4': 'TLS_DHE_PSK_WITH_NULL_SHA256',
    b'\x00\xB5': 'TLS_DHE_PSK_WITH_NULL_SHA384',
    b'\x00\xB6': 'TLS_RSA_PSK_WITH_AES_128_CBC_SHA256',
    b'\x00\xB7': 'TLS_RSA_PSK_WITH_AES_256_CBC_SHA384',
    b'\x00\xB8': 'TLS_RSA_PSK_WITH_NULL_SHA256',
    b'\x00\xB9': 'TLS_RSA_PSK_WITH_NULL_SHA384',
    b'\x00\xBA': 'TLS_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBB': 'TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBC': 'TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBD': 'TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBE': 'TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xBF': 'TLS_DH_anon_WITH_CAMELLIA_128_CBC_SHA256',
    b'\x00\xC0': 'TLS_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC1': 'TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC2': 'TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC3': 'TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC4': 'TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xC5': 'TLS_DH_anon_WITH_CAMELLIA_256_CBC_SHA256',
    b'\x00\xFF': 'TLS_EMPTY_RENEGOTIATION_INFO_SCSV',
    b'\x13\x01': 'TLS_AES_128_GCM_SHA256',
    b'\x13\x02': 'TLS_AES_256_GCM_SHA384',
    b'\x13\x03': 'TLS_CHACHA20_POLY1305_SHA256',
    b'\x13\x04': 'TLS_AES_128_CCM_SHA256',
    b'\x13\x05': 'TLS_AES_128_CCM_8_SHA256',
    b'\xC0\x01': 'TLS_ECDH_ECDSA_WITH_NULL_SHA',
    b'\xC0\x02': 'TLS_ECDH_ECDSA_WITH_RC4_128_SHA',
    b'\xC0\x03': 'TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x04': 'TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x05': 'TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x06': 'TLS_ECDHE_ECDSA_WITH_NULL_SHA',
    b'\xC0\x07': 'TLS_ECDHE_ECDSA_WITH_RC4_128_SHA',
    b'\xC0\x08': 'TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x09': 'TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x0A': 'TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x0B': 'TLS_ECDH_RSA_WITH_NULL_SHA',
    b'\xC0\x0C': 'TLS_ECDH_RSA_WITH_RC4_128_SHA',
    b'\xC0\x0D': 'TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x0E': 'TLS_ECDH_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x0F': 'TLS_ECDH_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x10': 'TLS_ECDHE_RSA_WITH_NULL_SHA',
    b'\xC0\x11': 'TLS_ECDHE_RSA_WITH_RC4_128_SHA',
    b'\xC0\x12': 'TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x13': 'TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x14': 'TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x15': 'TLS_ECDH_anon_WITH_NULL_SHA',
    b'\xC0\x16': 'TLS_ECDH_anon_WITH_RC4_128_SHA',
    b'\xC0\x17': 'TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x18': 'TLS_ECDH_anon_WITH_AES_128_CBC_SHA',
    b'\xC0\x19': 'TLS_ECDH_anon_WITH_AES_256_CBC_SHA',
    b'\xC0\x1A': 'TLS_SRP_SHA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1B': 'TLS_SRP_SHA_RSA_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1C': 'TLS_SRP_SHA_DSS_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x1D': 'TLS_SRP_SHA_WITH_AES_128_CBC_SHA',
    b'\xC0\x1E': 'TLS_SRP_SHA_RSA_WITH_AES_128_CBC_SHA',
    b'\xC0\x1F': 'TLS_SRP_SHA_DSS_WITH_AES_128_CBC_SHA',
    b'\xC0\x20': 'TLS_SRP_SHA_WITH_AES_256_CBC_SHA',
    b'\xC0\x21': 'TLS_SRP_SHA_RSA_WITH_AES_256_CBC_SHA',
    b'\xC0\x22': 'TLS_SRP_SHA_DSS_WITH_AES_256_CBC_SHA',
    b'\xC0\x23': 'TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x24': 'TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x25': 'TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x26': 'TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x27': 'TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x28': 'TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x29': 'TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256',
    b'\xC0\x2A': 'TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384',
    b'\xC0\x2B': 'TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2C': 'TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x2D': 'TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x2E': 'TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x2F': 'TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x30': 'TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x31': 'TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256',
    b'\xC0\x32': 'TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384',
    b'\xC0\x33': 'TLS_ECDHE_PSK_WITH_RC4_128_SHA',
    b'\xC0\x34': 'TLS_ECDHE_PSK_WITH_3DES_EDE_CBC_SHA',
    b'\xC0\x35': 'TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA',
    b'\xC0\x36': 'TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA',
    b'\xC0\x37': 'TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA256',
    b'\xC0\x38': 'TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA384',
    b'\xC0\x39': 'TLS_ECDHE_PSK_WITH_NULL_SHA',
    b'\xC0\x3A': 'TLS_ECDHE_PSK_WITH_NULL_SHA256',
    b'\xC0\x3B': 'TLS_ECDHE_PSK_WITH_NULL_SHA384',
    b'\xC0\x3C': 'TLS_RSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x3D': 'TLS_RSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x3E': 'TLS_DH_DSS_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x3F': 'TLS_DH_DSS_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x40': 'TLS_DH_RSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x41': 'TLS_DH_RSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x42': 'TLS_DHE_DSS_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x43': 'TLS_DHE_DSS_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x44': 'TLS_DHE_RSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x45': 'TLS_DHE_RSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x46': 'TLS_DH_anon_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x47': 'TLS_DH_anon_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x48': 'TLS_ECDHE_ECDSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x49': 'TLS_ECDHE_ECDSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x4A': 'TLS_ECDH_ECDSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x4B': 'TLS_ECDH_ECDSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x4C': 'TLS_ECDHE_RSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x4D': 'TLS_ECDHE_RSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x4E': 'TLS_ECDH_RSA_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x4F': 'TLS_ECDH_RSA_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x50': 'TLS_RSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x51': 'TLS_RSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x52': 'TLS_DHE_RSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x53': 'TLS_DHE_RSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x54': 'TLS_DH_RSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x55': 'TLS_DH_RSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x56': 'TLS_DHE_DSS_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x57': 'TLS_DHE_DSS_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x58': 'TLS_DH_DSS_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x59': 'TLS_DH_DSS_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x5A': 'TLS_DH_anon_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x5B': 'TLS_DH_anon_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x5C': 'TLS_ECDHE_ECDSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x5D': 'TLS_ECDHE_ECDSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x5E': 'TLS_ECDH_ECDSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x5F': 'TLS_ECDH_ECDSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x60': 'TLS_ECDHE_RSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x61': 'TLS_ECDHE_RSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x62': 'TLS_ECDH_RSA_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x63': 'TLS_ECDH_RSA_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x64': 'TLS_PSK_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x65': 'TLS_PSK_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x66': 'TLS_DHE_PSK_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x67': 'TLS_DHE_PSK_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x68': 'TLS_RSA_PSK_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x69': 'TLS_RSA_PSK_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x6A': 'TLS_PSK_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x6B': 'TLS_PSK_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x6C': 'TLS_DHE_PSK_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x6D': 'TLS_DHE_PSK_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x6E': 'TLS_RSA_PSK_WITH_ARIA_128_GCM_SHA256',
    b'\xC0\x6F': 'TLS_RSA_PSK_WITH_ARIA_256_GCM_SHA384',
    b'\xC0\x70': 'TLS_ECDHE_PSK_WITH_ARIA_128_CBC_SHA256',
    b'\xC0\x71': 'TLS_ECDHE_PSK_WITH_ARIA_256_CBC_SHA384',
    b'\xC0\x72': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x73': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x74': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x75': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x76': 'TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x77': 'TLS_ECDHE_RSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x78': 'TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x79': 'TLS_ECDH_RSA_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x7A': 'TLS_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x7B': 'TLS_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x7C': 'TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x7D': 'TLS_DHE_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x7E': 'TLS_DH_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x7F': 'TLS_DH_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x80': 'TLS_DHE_DSS_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x81': 'TLS_DHE_DSS_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x82': 'TLS_DH_DSS_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x83': 'TLS_DH_DSS_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x84': 'TLS_DH_anon_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x85': 'TLS_DH_anon_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x86': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x87': 'TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x88': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x89': 'TLS_ECDH_ECDSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x8A': 'TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x8B': 'TLS_ECDHE_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x8C': 'TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x8D': 'TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x8E': 'TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x8F': 'TLS_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x90': 'TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x91': 'TLS_DHE_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x92': 'TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256',
    b'\xC0\x93': 'TLS_RSA_PSK_WITH_CAMELLIA_256_GCM_SHA384',
    b'\xC0\x94': 'TLS_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x95': 'TLS_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x96': 'TLS_DHE_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x97': 'TLS_DHE_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x98': 'TLS_RSA_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x99': 'TLS_RSA_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x9A': 'TLS_ECDHE_PSK_WITH_CAMELLIA_128_CBC_SHA256',
    b'\xC0\x9B': 'TLS_ECDHE_PSK_WITH_CAMELLIA_256_CBC_SHA384',
    b'\xC0\x9C': 'TLS_RSA_WITH_AES_128_CCM',
    b'\xC0\x9D': 'TLS_RSA_WITH_AES_256_CCM',
    b'\xC0\x9E': 'TLS_DHE_RSA_WITH_AES_128_CCM',
    b'\xC0\x9F': 'TLS_DHE_RSA_WITH_AES_256_CCM',
    b'\xC0\xA0': 'TLS_RSA_WITH_AES_128_CCM_8',
    b'\xC0\xA1': 'TLS_RSA_WITH_AES_256_CCM_8',
    b'\xC0\xA2': 'TLS_DHE_RSA_WITH_AES_128_CCM_8',
    b'\xC0\xA3': 'TLS_DHE_RSA_WITH_AES_256_CCM_8',
    b'\xC0\xA4': 'TLS_PSK_WITH_AES_128_CCM',
    b'\xC0\xA5': 'TLS_PSK_WITH_AES_256_CCM',
    b'\xC0\xA6': 'TLS_DHE_PSK_WITH_AES_128_CCM',
    b'\xC0\xA7': 'TLS_DHE_PSK_WITH_AES_256_CCM',
    b'\xC0\xA8': 'TLS_PSK_WITH_AES_128_CCM_8',
    b'\xC0\xA9': 'TLS_PSK_WITH_AES_256_CCM_8',
    b'\xC0\xAA': 'TLS_PSK_DHE_WITH_AES_128_CCM_8',
    b'\xC0\xAB': 'TLS_PSK_DHE_WITH_AES_256_CCM_8',
    b'\xCC\xA8': 'TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xA9': 'TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAA': 'TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAB': 'TLS_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAC': 'TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAD': 'TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256',
    b'\xCC\xAE': 'TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256',
}



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import re

from ctypes import CDLL, c_void_p, c_char_p, c_int, c_ulong, c_uint, c_long, c_size_t, POINTER

from .. import _backend_config
from .._errors import pretty_message
from .._ffi import FFIEngineError, get_library
from ..errors import LibraryNotFoundError


__all__ = [
    'is_libressl',
    'libcrypto',
    'libressl_version',
    'libressl_version_info',
    'version',
    'version_info',
]


libcrypto_path = _backend_config().get('libcrypto_path')
if libcrypto_path is None:
    libcrypto_path = get_library('crypto', 'libcrypto.dylib', '42')
if not libcrypto_path:
    raise LibraryNotFoundError('The library libcrypto could not be found')

libcrypto = CDLL(libcrypto_path, use_errno=True)

try:
    libcrypto.SSLeay_version.argtypes = [c_int]
    libcrypto.SSLeay_version.restype = c_char_p
    version_string = libcrypto.SSLeay_version(0).decode('utf-8')
except (AttributeError):
    libcrypto.OpenSSL_version.argtypes = [c_int]
    libcrypto.OpenSSL_version.restype = c_char_p
    version_string = libcrypto.OpenSSL_version(0).decode('utf-8')

is_libressl = 'LibreSSL' in version_string

version_match = re.search('\\b(\\d\\.\\d\\.\\d[a-z]*)\\b', version_string)
if not version_match:
    version_match = re.search('(?<=LibreSSL )(\\d\\.\\d(\\.\\d)?)\\b', version_string)
if not version_match:
    raise LibraryNotFoundError('Error detecting the version of libcrypto')
version = version_match.group(1)
version_parts = re.sub('(\\d)([a-z]+)', '\\1.\\2', version).split('.')
version_info = tuple(int(part) if part.isdigit() else part for part in version_parts)

# LibreSSL is compatible with libcrypto from OpenSSL 1.0.1
libressl_version = ''
libressl_version_info = tuple()
if is_libressl:
    libressl_version = version
    libressl_version_info = version_info
    version = '1.0.1'
    version_info = (1, 0, 1)

if version_info < (0, 9, 8):
    raise LibraryNotFoundError(pretty_message(
        '''
        OpenSSL versions older than 0.9.8 are not supported - found version %s
        ''',
        version
    ))

P_EVP_CIPHER_CTX = c_void_p
P_EVP_CIPHER = c_void_p

P_EVP_MD_CTX = c_void_p
P_EVP_MD = c_void_p

P_ENGINE = c_void_p
OSSL_PROVIDER = c_void_p
OSSL_LIB_CTX = c_void_p

P_EVP_PKEY = c_void_p
EVP_PKEY_CTX = c_void_p
P_EVP_PKEY_CTX = POINTER(c_void_p)
P_X509 = POINTER(c_void_p)
P_DH = c_void_p
P_RSA = c_void_p
P_DSA = c_void_p
P_EC_KEY = c_void_p
P_BN_GENCB = c_void_p
BIGNUM = c_void_p
P_BIGNUM = POINTER(BIGNUM)

p_int = POINTER(c_int)
p_uint = POINTER(c_uint)

try:
    if version_info < (1, 1):
        libcrypto.ERR_load_crypto_strings.argtypes = []
        libcrypto.ERR_load_crypto_strings.restype = None

        libcrypto.ERR_free_strings.argtypes = []
        libcrypto.ERR_free_strings.restype = None

    if version_info >= (3, ):
        libcrypto.OSSL_PROVIDER_available.argtypes = [OSSL_LIB_CTX, c_char_p]
        libcrypto.OSSL_PROVIDER_available.restype = c_int

        libcrypto.OSSL_PROVIDER_load.argtypes = [OSSL_LIB_CTX, c_char_p]
        libcrypto.OSSL_PROVIDER_load.restype = POINTER(OSSL_PROVIDER)

    libcrypto.ERR_get_error.argtypes = []
    libcrypto.ERR_get_error.restype = c_ulong

    libcrypto.ERR_peek_error.argtypes = []
    libcrypto.ERR_peek_error.restype = c_ulong

    libcrypto.ERR_error_string.argtypes = [
        c_ulong,
        c_char_p
    ]
    libcrypto.ERR_error_string.restype = c_char_p

    libcrypto.OPENSSL_config.argtypes = [
        c_char_p
    ]
    libcrypto.OPENSSL_config.restype = None

    # This allocates the memory and inits
    libcrypto.EVP_CIPHER_CTX_new.argtype = []
    libcrypto.EVP_CIPHER_CTX_new.restype = P_EVP_CIPHER_CTX

    libcrypto.EVP_CIPHER_CTX_set_key_length.argtypes = [
        P_EVP_CIPHER_CTX,
        c_int
    ]
    libcrypto.EVP_CIPHER_CTX_set_key_length.restype = c_int

    libcrypto.EVP_CIPHER_CTX_set_padding.argtypes = [
        P_EVP_CIPHER_CTX,
        c_int
    ]
    libcrypto.EVP_CIPHER_CTX_set_padding.restype = c_int

    libcrypto.EVP_CIPHER_CTX_ctrl.argtypes = [
        P_EVP_CIPHER_CTX,
        c_int,
        c_int,
        c_void_p
    ]
    libcrypto.EVP_CIPHER_CTX_ctrl.restype = c_int

    # This cleans up and frees
    libcrypto.EVP_CIPHER_CTX_free.argtypes = [
        P_EVP_CIPHER_CTX
    ]
    libcrypto.EVP_CIPHER_CTX_free.restype = None

    libcrypto.EVP_aes_128_cbc.argtypes = []
    libcrypto.EVP_aes_128_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_aes_192_cbc.argtypes = []
    libcrypto.EVP_aes_192_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_aes_256_cbc.argtypes = []
    libcrypto.EVP_aes_256_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_des_cbc.argtypes = []
    libcrypto.EVP_des_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_des_ede_cbc.argtypes = []
    libcrypto.EVP_des_ede_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_des_ede3_cbc.argtypes = []
    libcrypto.EVP_des_ede3_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_rc4.argtypes = []
    libcrypto.EVP_rc4.restype = P_EVP_CIPHER

    libcrypto.EVP_rc2_cbc.argtypes = []
    libcrypto.EVP_rc2_cbc.restype = P_EVP_CIPHER

    libcrypto.EVP_EncryptInit_ex.argtypes = [
        P_EVP_CIPHER_CTX,
        P_EVP_CIPHER,
        P_ENGINE,
        c_char_p,
        c_char_p
    ]
    libcrypto.EVP_EncryptInit_ex.restype = c_int

    libcrypto.EVP_EncryptUpdate.argtypes = [
        P_EVP_CIPHER_CTX,
        c_char_p,
        p_int,
        c_char_p,
        c_int
    ]
    libcrypto.EVP_EncryptUpdate.restype = c_int

    libcrypto.EVP_EncryptFinal_ex.argtypes = [
        P_EVP_CIPHER_CTX,
        c_char_p,
        p_int
    ]
    libcrypto.EVP_EncryptFinal_ex.restype = c_int

    libcrypto.EVP_DecryptInit_ex.argtypes = [
        P_EVP_CIPHER_CTX,
        P_EVP_CIPHER,
        P_ENGINE,
        c_char_p,
        c_char_p
    ]
    libcrypto.EVP_DecryptInit_ex.restype = c_int

    libcrypto.EVP_DecryptUpdate.argtypes = [
        P_EVP_CIPHER_CTX,
        c_char_p,
        p_int,
        c_char_p,
        c_int
    ]
    libcrypto.EVP_DecryptUpdate.restype = c_int

    libcrypto.EVP_DecryptFinal_ex.argtypes = [
        P_EVP_CIPHER_CTX,
        c_char_p,
        p_int
    ]
    libcrypto.EVP_DecryptFinal_ex.restype = c_int

    libcrypto.d2i_AutoPrivateKey.argtypes = [
        POINTER(P_EVP_PKEY),
        POINTER(c_char_p),
        c_int
    ]
    libcrypto.d2i_AutoPrivateKey.restype = P_EVP_PKEY

    libcrypto.d2i_PUBKEY.argtypes = [
        POINTER(P_EVP_PKEY),
        POINTER(c_char_p),
        c_int
    ]
    libcrypto.d2i_PUBKEY.restype = P_EVP_PKEY

    libcrypto.i2d_PUBKEY.argtypes = [
        P_EVP_PKEY,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_PUBKEY.restype = c_int

    libcrypto.d2i_X509.argtypes = [
        POINTER(P_X509),
        POINTER(c_char_p),
        c_int
    ]
    libcrypto.d2i_X509.restype = P_X509

    libcrypto.i2d_X509.argtypes = [
        P_X509,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_X509.restype = c_int

    libcrypto.X509_get_pubkey.argtypes = [
        P_X509
    ]
    libcrypto.X509_get_pubkey.restype = P_EVP_PKEY

    libcrypto.X509_free.argtypes = [
        P_X509
    ]
    libcrypto.X509_free.restype = None

    libcrypto.EVP_PKEY_free.argtypes = [
        P_EVP_PKEY
    ]
    libcrypto.EVP_PKEY_free.restype = None

    if version_info < (1, 1):
        libcrypto.EVP_MD_CTX_create.argtypes = []
        libcrypto.EVP_MD_CTX_create.restype = P_EVP_MD_CTX

        libcrypto.EVP_MD_CTX_destroy.argtypes = [
            P_EVP_MD_CTX
        ]
        libcrypto.EVP_MD_CTX_destroy.restype = None
    else:
        libcrypto.EVP_MD_CTX_new.argtypes = []
        libcrypto.EVP_MD_CTX_new.restype = P_EVP_MD_CTX

        libcrypto.EVP_MD_CTX_free.argtypes = [
            P_EVP_MD_CTX
        ]
        libcrypto.EVP_MD_CTX_free.restype = None

    libcrypto.EVP_md5.argtypes = []
    libcrypto.EVP_md5.restype = P_EVP_MD

    libcrypto.EVP_sha1.argtypes = []
    libcrypto.EVP_sha1.restype = P_EVP_MD

    libcrypto.EVP_sha224.argtypes = []
    libcrypto.EVP_sha224.restype = P_EVP_MD

    libcrypto.EVP_sha256.argtypes = []
    libcrypto.EVP_sha256.restype = P_EVP_MD

    libcrypto.EVP_sha384.argtypes = []
    libcrypto.EVP_sha384.restype = P_EVP_MD

    libcrypto.EVP_sha512.argtypes = []
    libcrypto.EVP_sha512.restype = P_EVP_MD

    if version_info < (3, 0):
        libcrypto.EVP_PKEY_size.argtypes = [
            P_EVP_PKEY
        ]
        libcrypto.EVP_PKEY_size.restype = c_int
    else:
        libcrypto.EVP_PKEY_get_size.argtypes = [
            P_EVP_PKEY
        ]
        libcrypto.EVP_PKEY_get_size.restype = c_int

    libcrypto.EVP_PKEY_get1_RSA.argtypes = [
        P_EVP_PKEY
    ]
    libcrypto.EVP_PKEY_get1_RSA.restype = P_RSA

    libcrypto.RSA_free.argtypes = [
        P_RSA
    ]
    libcrypto.RSA_free.restype = None

    libcrypto.RSA_public_encrypt.argtypes = [
        c_int,
        c_char_p,
        c_char_p,
        P_RSA,
        c_int
    ]
    libcrypto.RSA_public_encrypt.restype = c_int

    libcrypto.RSA_private_encrypt.argtypes = [
        c_int,
        c_char_p,
        c_char_p,
        P_RSA,
        c_int
    ]
    libcrypto.RSA_private_encrypt.restype = c_int

    libcrypto.RSA_public_decrypt.argtypes = [
        c_int,
        c_char_p,
        c_char_p,
        P_RSA,
        c_int
    ]
    libcrypto.RSA_public_decrypt.restype = c_int

    libcrypto.RSA_private_decrypt.argtypes = [
        c_int,
        c_char_p,
        c_char_p,
        P_RSA,
        c_int
    ]
    libcrypto.RSA_private_decrypt.restype = c_int

    libcrypto.EVP_DigestUpdate.argtypes = [
        P_EVP_MD_CTX,
        c_char_p,
        c_uint
    ]
    libcrypto.EVP_DigestUpdate.restype = c_int

    libcrypto.PKCS12_key_gen_uni.argtypes = [
        c_char_p,
        c_int,
        c_char_p,
        c_int,
        c_int,
        c_int,
        c_int,
        c_char_p,
        c_void_p
    ]
    libcrypto.PKCS12_key_gen_uni.restype = c_int

    libcrypto.BN_free.argtypes = [
        P_BIGNUM
    ]
    libcrypto.BN_free.restype = None

    libcrypto.BN_dec2bn.argtypes = [
        POINTER(P_BIGNUM),
        c_char_p
    ]
    libcrypto.BN_dec2bn.restype = c_int

    libcrypto.DH_new.argtypes = []
    libcrypto.DH_new.restype = P_DH

    libcrypto.DH_generate_parameters_ex.argtypes = [
        P_DH,
        c_int,
        c_int,
        P_BN_GENCB
    ]
    libcrypto.DH_generate_parameters_ex.restype = c_int

    libcrypto.i2d_DHparams.argtypes = [
        P_DH,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_DHparams.restype = c_int

    libcrypto.DH_free.argtypes = [
        P_DH
    ]
    libcrypto.DH_free.restype = None

    libcrypto.RSA_new.argtypes = []
    libcrypto.RSA_new.restype = P_RSA

    libcrypto.RSA_generate_key_ex.argtypes = [
        P_RSA,
        c_int,
        P_BIGNUM,
        P_BN_GENCB
    ]
    libcrypto.RSA_generate_key_ex.restype = c_int

    libcrypto.i2d_RSAPublicKey.argtypes = [
        P_RSA,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_RSAPublicKey.restype = c_int

    libcrypto.i2d_RSAPrivateKey.argtypes = [
        P_RSA,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_RSAPrivateKey.restype = c_int

    libcrypto.RSA_free.argtypes = [
        P_RSA
    ]
    libcrypto.RSA_free.restype = None

    libcrypto.DSA_new.argtypes = []
    libcrypto.DSA_new.restype = P_DSA

    libcrypto.DSA_generate_parameters_ex.argtypes = [
        P_DSA,
        c_int,
        c_char_p,
        c_int,
        POINTER(c_int),
        POINTER(c_ulong),
        P_BN_GENCB
    ]
    libcrypto.DSA_generate_parameters_ex.restype = c_int

    libcrypto.DSA_generate_key.argtypes = [
        P_DSA
    ]
    libcrypto.DSA_generate_key.restype = c_int

    libcrypto.i2d_DSA_PUBKEY.argtypes = [
        P_DSA,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_DSA_PUBKEY.restype = c_int

    libcrypto.i2d_DSAPrivateKey.argtypes = [
        P_DSA,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_DSAPrivateKey.restype = c_int

    libcrypto.DSA_free.argtypes = [
        P_DSA
    ]
    libcrypto.DSA_free.restype = None

    libcrypto.EC_KEY_new_by_curve_name.argtypes = [
        c_int
    ]
    libcrypto.EC_KEY_new_by_curve_name.restype = P_EC_KEY

    libcrypto.EC_KEY_generate_key.argtypes = [
        P_EC_KEY
    ]
    libcrypto.EC_KEY_generate_key.restype = c_int

    libcrypto.EC_KEY_set_asn1_flag.argtypes = [
        P_EC_KEY,
        c_int
    ]
    libcrypto.EC_KEY_set_asn1_flag.restype = None

    libcrypto.i2d_ECPrivateKey.argtypes = [
        P_EC_KEY,
        POINTER(c_char_p)
    ]
    libcrypto.i2d_ECPrivateKey.restype = c_int

    libcrypto.i2o_ECPublicKey.argtypes = [
        P_EC_KEY,
        POINTER(c_char_p)
    ]
    libcrypto.i2o_ECPublicKey.restype = c_int

    libcrypto.EC_KEY_free.argtypes = [
        P_EC_KEY
    ]
    libcrypto.EC_KEY_free.restype = None

    if version_info < (1,):
        P_DSA_SIG = c_void_p
        P_ECDSA_SIG = c_void_p

        libcrypto.DSA_do_sign.argtypes = [
            c_char_p,
            c_int,
            P_DSA
        ]
        libcrypto.DSA_do_sign.restype = P_DSA_SIG

        libcrypto.ECDSA_do_sign.argtypes = [
            c_char_p,
            c_int,
            P_EC_KEY
        ]
        libcrypto.ECDSA_do_sign.restype = P_ECDSA_SIG

        libcrypto.d2i_DSA_SIG.argtypes = [
            POINTER(P_DSA_SIG),
            POINTER(c_char_p),
            c_long
        ]
        libcrypto.d2i_DSA_SIG.restype = P_DSA_SIG

        libcrypto.d2i_ECDSA_SIG.argtypes = [
            POINTER(P_ECDSA_SIG),
            POINTER(c_char_p),
            c_long
        ]
        libcrypto.d2i_ECDSA_SIG.restype = P_ECDSA_SIG

        libcrypto.i2d_DSA_SIG.argtypes = [
            P_DSA_SIG,
            POINTER(c_char_p)
        ]
        libcrypto.i2d_DSA_SIG.restype = c_int

        libcrypto.i2d_ECDSA_SIG.argtypes = [
            P_ECDSA_SIG,
            POINTER(c_char_p)
        ]
        libcrypto.i2d_ECDSA_SIG.restype = c_int

        libcrypto.DSA_do_verify.argtypes = [
            c_char_p,
            c_int,
            P_DSA_SIG,
            P_DSA
        ]
        libcrypto.DSA_do_verify.restype = c_int

        libcrypto.ECDSA_do_verify.argtypes = [
            c_char_p,
            c_int,
            P_ECDSA_SIG,
            P_EC_KEY
        ]
        libcrypto.ECDSA_do_verify.restype = c_int

        libcrypto.DSA_SIG_free.argtypes = [
            P_DSA_SIG
        ]
        libcrypto.DSA_SIG_free.restype = None

        libcrypto.ECDSA_SIG_free.argtypes = [
            P_ECDSA_SIG
        ]
        libcrypto.ECDSA_SIG_free.restype = None

        libcrypto.EVP_PKEY_get1_DSA.argtypes = [
            P_EVP_PKEY
        ]
        libcrypto.EVP_PKEY_get1_DSA.restype = P_DSA

        libcrypto.EVP_PKEY_get1_EC_KEY.argtypes = [
            P_EVP_PKEY
        ]
        libcrypto.EVP_PKEY_get1_EC_KEY.restype = P_EC_KEY

        libcrypto.RSA_verify_PKCS1_PSS.argtypes = [
            P_RSA,
            c_char_p,
            P_EVP_MD,
            c_char_p,
            c_int
        ]
        libcrypto.RSA_verify_PKCS1_PSS.restype = c_int

        libcrypto.RSA_padding_add_PKCS1_PSS.argtypes = [
            P_RSA,
            c_char_p,
            c_char_p,
            P_EVP_MD,
            c_int
        ]
        libcrypto.RSA_padding_add_PKCS1_PSS.restype = c_int

        libcrypto.EVP_DigestInit_ex.argtypes = [
            P_EVP_MD_CTX,
            P_EVP_MD,
            P_ENGINE
        ]
        libcrypto.EVP_DigestInit_ex.restype = c_int

        libcrypto.EVP_SignFinal.argtypes = [
            P_EVP_MD_CTX,
            c_char_p,
            p_uint,
            P_EVP_PKEY
        ]
        libcrypto.EVP_SignFinal.restype = c_int

        libcrypto.EVP_VerifyFinal.argtypes = [
            P_EVP_MD_CTX,
            c_char_p,
            c_uint,
            P_EVP_PKEY
        ]
        libcrypto.EVP_VerifyFinal.restype = c_int

        libcrypto.EVP_MD_CTX_set_flags.argtypes = [
            P_EVP_MD_CTX,
            c_int
        ]
        libcrypto.EVP_MD_CTX_set_flags.restype = None

    else:
        libcrypto.PKCS5_PBKDF2_HMAC.argtypes = [
            c_char_p,
            c_int,
            c_char_p,
            c_int,
            c_int,
            P_EVP_MD,
            c_int,
            c_char_p
        ]
        libcrypto.PKCS5_PBKDF2_HMAC.restype = c_int

        libcrypto.EVP_DigestSignInit.argtypes = [
            P_EVP_MD_CTX,
            POINTER(P_EVP_PKEY_CTX),
            P_EVP_MD,
            P_ENGINE,
            P_EVP_PKEY
        ]
        libcrypto.EVP_DigestSignInit.restype = c_int

        libcrypto.EVP_DigestSignFinal.argtypes = [
            P_EVP_MD_CTX,
            c_char_p,
            POINTER(c_size_t)
        ]
        libcrypto.EVP_DigestSignFinal.restype = c_int

        libcrypto.EVP_DigestVerifyInit.argtypes = [
            P_EVP_MD_CTX,
            POINTER(P_EVP_PKEY_CTX),
            P_EVP_MD,
            P_ENGINE,
            P_EVP_PKEY
        ]
        libcrypto.EVP_DigestVerifyInit.restype = c_int

        libcrypto.EVP_DigestVerifyFinal.argtypes = [
            P_EVP_MD_CTX,
            c_char_p,
            c_size_t
        ]
        libcrypto.EVP_DigestVerifyFinal.restype = c_int

        libcrypto.EVP_PKEY_CTX_ctrl.argtypes = [
            P_EVP_PKEY_CTX,
            c_int,
            c_int,
            c_int,
            c_int,
            c_void_p
        ]
        libcrypto.EVP_PKEY_CTX_ctrl.restype = c_int

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')


setattr(libcrypto, 'EVP_PKEY_CTX', EVP_PKEY_CTX)
setattr(libcrypto, 'BIGNUM', BIGNUM)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .._errors import pretty_message
from .._ffi import (
    buffer_from_bytes,
    bytes_from_buffer,
    deref,
    new,
    null,
    pointer_set,
    struct,
    struct_bytes,
    unwrap,
    write_to_buffer,
)
from .util import rand_bytes
from .. import backend
from .._types import type_name, byte_cls

_backend = backend()

if _backend == 'winlegacy':
    from ._advapi32 import advapi32, Advapi32Const, handle_error, open_context_handle, close_context_handle
else:
    from ._cng import bcrypt, BcryptConst, handle_error, open_alg_handle, close_alg_handle


__all__ = [
    'aes_cbc_no_padding_decrypt',
    'aes_cbc_no_padding_encrypt',
    'aes_cbc_pkcs7_decrypt',
    'aes_cbc_pkcs7_encrypt',
    'des_cbc_pkcs5_decrypt',
    'des_cbc_pkcs5_encrypt',
    'rc2_cbc_pkcs5_decrypt',
    'rc2_cbc_pkcs5_encrypt',
    'rc4_decrypt',
    'rc4_encrypt',
    'tripledes_cbc_pkcs5_decrypt',
    'tripledes_cbc_pkcs5_encrypt',
]


def aes_cbc_no_padding_encrypt(key, data, iv):
    """
    Encrypts plaintext using AES in CBC mode with a 128, 192 or 256 bit key and
    no padding. This means the ciphertext must be an exact multiple of 16 bytes
    long.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - either a byte string 16-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) not in [16, 24, 32]:
        raise ValueError(pretty_message(
            '''
            key must be either 16, 24 or 32 bytes (128, 192 or 256 bits)
            long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(16)
    elif len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    if len(data) % 16 != 0:
        raise ValueError(pretty_message(
            '''
            data must be a multiple of 16 bytes long - is %s
            ''',
            len(data)
        ))

    return (iv, _encrypt('aes', key, data, iv, False))


def aes_cbc_no_padding_decrypt(key, data, iv):
    """
    Decrypts AES ciphertext in CBC mode using a 128, 192 or 256 bit key and no
    padding.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 16-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) not in [16, 24, 32]:
        raise ValueError(pretty_message(
            '''
            key must be either 16, 24 or 32 bytes (128, 192 or 256 bits)
            long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('aes', key, data, iv, False)


def aes_cbc_pkcs7_encrypt(key, data, iv):
    """
    Encrypts plaintext using AES in CBC mode with a 128, 192 or 256 bit key and
    PKCS#7 padding.

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - either a byte string 16-bytes long or None
        to generate an IV

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) not in [16, 24, 32]:
        raise ValueError(pretty_message(
            '''
            key must be either 16, 24 or 32 bytes (128, 192 or 256 bits)
            long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(16)
    elif len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt('aes', key, data, iv, True))


def aes_cbc_pkcs7_decrypt(key, data, iv):
    """
    Decrypts AES ciphertext in CBC mode using a 128, 192 or 256 bit key

    :param key:
        The encryption key - a byte string either 16, 24 or 32 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string 16-bytes long

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) not in [16, 24, 32]:
        raise ValueError(pretty_message(
            '''
            key must be either 16, 24 or 32 bytes (128, 192 or 256 bits)
            long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 16:
        raise ValueError(pretty_message(
            '''
            iv must be 16 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('aes', key, data, iv, True)


def rc4_encrypt(key, data):
    """
    Encrypts plaintext using RC4 with a 40-128 bit key

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The plaintext - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the ciphertext
    """

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    return _encrypt('rc4', key, data, None, None)


def rc4_decrypt(key, data):
    """
    Decrypts RC4 ciphertext using a 40-128 bit key

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The ciphertext - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    return _decrypt('rc4', key, data, None, None)


def rc2_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using RC2 with a 64 bit key

    :param key:
        The encryption key - a byte string 8 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The 8-byte initialization vector to use - a byte string - set as None
        to generate an appropriate one

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt('rc2', key, data, iv, True))


def rc2_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts RC2 ciphertext using a 64 bit key

    :param key:
        The encryption key - a byte string 8 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector used for encryption - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) < 5 or len(key) > 16:
        raise ValueError(pretty_message(
            '''
            key must be 5 to 16 bytes (40 to 128 bits) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('rc2', key, data, iv, True)


def tripledes_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using 3DES in either 2 or 3 key mode

    :param key:
        The encryption key - a byte string 16 or 24 bytes long (2 or 3 key mode)

    :param data:
        The plaintext - a byte string

    :param iv:
        The 8-byte initialization vector to use - a byte string - set as None
        to generate an appropriate one

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) != 16 and len(key) != 24:
        raise ValueError(pretty_message(
            '''
            key must be 16 bytes (2 key) or 24 bytes (3 key) long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    cipher = 'tripledes_3key'
    if len(key) == 16:
        cipher = 'tripledes_2key'

    return (iv, _encrypt(cipher, key, data, iv, True))


def tripledes_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts 3DES ciphertext in either 2 or 3 key mode

    :param key:
        The encryption key - a byte string 16 or 24 bytes long (2 or 3 key mode)

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector used for encryption - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) != 16 and len(key) != 24:
        raise ValueError(pretty_message(
            '''
            key must be 16 bytes (2 key) or 24 bytes (3 key) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    cipher = 'tripledes_3key'
    if len(key) == 16:
        cipher = 'tripledes_2key'

    return _decrypt(cipher, key, data, iv, True)


def des_cbc_pkcs5_encrypt(key, data, iv):
    """
    Encrypts plaintext using DES with a 56 bit key

    :param key:
        The encryption key - a byte string 8 bytes long (includes error
        correction bits)

    :param data:
        The plaintext - a byte string

    :param iv:
        The 8-byte initialization vector to use - a byte string - set as None
        to generate an appropriate one

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A tuple of two byte strings (iv, ciphertext)
    """

    if len(key) != 8:
        raise ValueError(pretty_message(
            '''
            key must be 8 bytes (56 bits + 8 parity bits) long - is %s
            ''',
            len(key)
        ))

    if not iv:
        iv = rand_bytes(8)
    elif len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return (iv, _encrypt('des', key, data, iv, True))


def des_cbc_pkcs5_decrypt(key, data, iv):
    """
    Decrypts DES ciphertext using a 56 bit key

    :param key:
        The encryption key - a byte string 8 bytes long (includes error
        correction bits)

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector used for encryption - a byte string

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if len(key) != 8:
        raise ValueError(pretty_message(
            '''
            key must be 8 bytes (56 bits + 8 parity bits) long - is %s
            ''',
            len(key)
        ))

    if len(iv) != 8:
        raise ValueError(pretty_message(
            '''
            iv must be 8 bytes long - is %s
            ''',
            len(iv)
        ))

    return _decrypt('des', key, data, iv, True)


def _advapi32_create_handles(cipher, key, iv):
    """
    Creates an HCRYPTPROV and HCRYPTKEY for symmetric encryption/decryption. The
    HCRYPTPROV must be released by close_context_handle() and the
    HCRYPTKEY must be released by advapi32.CryptDestroyKey() when done.

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        A byte string of the symmetric key

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :return:
        A tuple of (HCRYPTPROV, HCRYPTKEY)
    """

    context_handle = None

    if cipher == 'aes':
        algorithm_id = {
            16: Advapi32Const.CALG_AES_128,
            24: Advapi32Const.CALG_AES_192,
            32: Advapi32Const.CALG_AES_256,
        }[len(key)]
    else:
        algorithm_id = {
            'des': Advapi32Const.CALG_DES,
            'tripledes_2key': Advapi32Const.CALG_3DES_112,
            'tripledes_3key': Advapi32Const.CALG_3DES,
            'rc2': Advapi32Const.CALG_RC2,
            'rc4': Advapi32Const.CALG_RC4,
        }[cipher]

    provider = Advapi32Const.MS_ENH_RSA_AES_PROV
    context_handle = open_context_handle(provider, verify_only=False)

    blob_header_pointer = struct(advapi32, 'BLOBHEADER')
    blob_header = unwrap(blob_header_pointer)
    blob_header.bType = Advapi32Const.PLAINTEXTKEYBLOB
    blob_header.bVersion = Advapi32Const.CUR_BLOB_VERSION
    blob_header.reserved = 0
    blob_header.aiKeyAlg = algorithm_id

    blob_struct_pointer = struct(advapi32, 'PLAINTEXTKEYBLOB')
    blob_struct = unwrap(blob_struct_pointer)
    blob_struct.hdr = blob_header
    blob_struct.dwKeySize = len(key)

    blob = struct_bytes(blob_struct_pointer) + key

    flags = 0
    if cipher in set(['rc2', 'rc4']) and len(key) == 5:
        flags = Advapi32Const.CRYPT_NO_SALT

    key_handle_pointer = new(advapi32, 'HCRYPTKEY *')
    res = advapi32.CryptImportKey(
        context_handle,
        blob,
        len(blob),
        null(),
        flags,
        key_handle_pointer
    )
    handle_error(res)

    key_handle = unwrap(key_handle_pointer)

    if cipher == 'rc2':
        buf = new(advapi32, 'DWORD *', len(key) * 8)
        res = advapi32.CryptSetKeyParam(
            key_handle,
            Advapi32Const.KP_EFFECTIVE_KEYLEN,
            buf,
            0
        )
        handle_error(res)

    if cipher != 'rc4':
        res = advapi32.CryptSetKeyParam(
            key_handle,
            Advapi32Const.KP_IV,
            iv,
            0
        )
        handle_error(res)

        buf = new(advapi32, 'DWORD *', Advapi32Const.CRYPT_MODE_CBC)
        res = advapi32.CryptSetKeyParam(
            key_handle,
            Advapi32Const.KP_MODE,
            buf,
            0
        )
        handle_error(res)

        buf = new(advapi32, 'DWORD *', Advapi32Const.PKCS5_PADDING)
        res = advapi32.CryptSetKeyParam(
            key_handle,
            Advapi32Const.KP_PADDING,
            buf,
            0
        )
        handle_error(res)

    return (context_handle, key_handle)


def _bcrypt_create_key_handle(cipher, key):
    """
    Creates a BCRYPT_KEY_HANDLE for symmetric encryption/decryption. The
    handle must be released by bcrypt.BCryptDestroyKey() when done.

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        A byte string of the symmetric key

    :return:
        A BCRYPT_KEY_HANDLE
    """

    alg_handle = None

    alg_constant = {
        'aes': BcryptConst.BCRYPT_AES_ALGORITHM,
        'des': BcryptConst.BCRYPT_DES_ALGORITHM,
        'tripledes_2key': BcryptConst.BCRYPT_3DES_112_ALGORITHM,
        'tripledes_3key': BcryptConst.BCRYPT_3DES_ALGORITHM,
        'rc2': BcryptConst.BCRYPT_RC2_ALGORITHM,
        'rc4': BcryptConst.BCRYPT_RC4_ALGORITHM,
    }[cipher]

    try:
        alg_handle = open_alg_handle(alg_constant)
        blob_type = BcryptConst.BCRYPT_KEY_DATA_BLOB

        blob_struct_pointer = struct(bcrypt, 'BCRYPT_KEY_DATA_BLOB_HEADER')
        blob_struct = unwrap(blob_struct_pointer)
        blob_struct.dwMagic = BcryptConst.BCRYPT_KEY_DATA_BLOB_MAGIC
        blob_struct.dwVersion = BcryptConst.BCRYPT_KEY_DATA_BLOB_VERSION1
        blob_struct.cbKeyData = len(key)

        blob = struct_bytes(blob_struct_pointer) + key

        if cipher == 'rc2':
            buf = new(bcrypt, 'DWORD *', len(key) * 8)
            res = bcrypt.BCryptSetProperty(
                alg_handle,
                BcryptConst.BCRYPT_EFFECTIVE_KEY_LENGTH,
                buf,
                4,
                0
            )
            handle_error(res)

        key_handle_pointer = new(bcrypt, 'BCRYPT_KEY_HANDLE *')
        res = bcrypt.BCryptImportKey(
            alg_handle,
            null(),
            blob_type,
            key_handle_pointer,
            null(),
            0,
            blob,
            len(blob),
            0
        )
        handle_error(res)

        return unwrap(key_handle_pointer)

    finally:
        if alg_handle:
            close_alg_handle(alg_handle)


def _encrypt(cipher, key, data, iv, padding):
    """
    Encrypts plaintext

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the ciphertext
    """

    if not isinstance(key, byte_cls):
        raise TypeError(pretty_message(
            '''
            key must be a byte string, not %s
            ''',
            type_name(key)
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if cipher != 'rc4' and not isinstance(iv, byte_cls):
        raise TypeError(pretty_message(
            '''
            iv must be a byte string, not %s
            ''',
            type_name(iv)
        ))

    if cipher != 'rc4' and not padding:
        # AES in CBC mode can be allowed with no padding if
        # the data is an exact multiple of the block size
        if not (cipher == 'aes' and len(data) % 16 == 0):
            raise ValueError('padding must be specified')

    if _backend == 'winlegacy':
        return _advapi32_encrypt(cipher, key, data, iv, padding)
    return _bcrypt_encrypt(cipher, key, data, iv, padding)


def _advapi32_encrypt(cipher, key, data, iv, padding):
    """
    Encrypts plaintext via CryptoAPI

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the ciphertext
    """

    context_handle = None
    key_handle = None

    try:
        context_handle, key_handle = _advapi32_create_handles(cipher, key, iv)

        out_len = new(advapi32, 'DWORD *', len(data))
        res = advapi32.CryptEncrypt(
            key_handle,
            null(),
            True,
            0,
            null(),
            out_len,
            0
        )
        handle_error(res)

        buffer_len = deref(out_len)
        buffer = buffer_from_bytes(buffer_len)
        write_to_buffer(buffer, data)

        pointer_set(out_len, len(data))
        res = advapi32.CryptEncrypt(
            key_handle,
            null(),
            True,
            0,
            buffer,
            out_len,
            buffer_len
        )
        handle_error(res)

        output = bytes_from_buffer(buffer, deref(out_len))

        # Remove padding when not required. CryptoAPI doesn't support this, so
        # we just manually remove it.
        if cipher == 'aes' and not padding and len(output) == len(data) + 16:
            output = output[:-16]

        return output

    finally:
        if key_handle:
            advapi32.CryptDestroyKey(key_handle)
        if context_handle:
            close_context_handle(context_handle)


def _bcrypt_encrypt(cipher, key, data, iv, padding):
    """
    Encrypts plaintext via CNG

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The plaintext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the ciphertext
    """

    key_handle = None

    try:
        key_handle = _bcrypt_create_key_handle(cipher, key)

        if iv is None:
            iv_len = 0
        else:
            iv_len = len(iv)

        flags = 0
        if padding is True:
            flags = BcryptConst.BCRYPT_BLOCK_PADDING

        out_len = new(bcrypt, 'ULONG *')
        res = bcrypt.BCryptEncrypt(
            key_handle,
            data,
            len(data),
            null(),
            null(),
            0,
            null(),
            0,
            out_len,
            flags
        )
        handle_error(res)

        buffer_len = deref(out_len)
        buffer = buffer_from_bytes(buffer_len)
        iv_buffer = buffer_from_bytes(iv) if iv else null()

        res = bcrypt.BCryptEncrypt(
            key_handle,
            data,
            len(data),
            null(),
            iv_buffer,
            iv_len,
            buffer,
            buffer_len,
            out_len,
            flags
        )
        handle_error(res)

        return bytes_from_buffer(buffer, deref(out_len))

    finally:
        if key_handle:
            bcrypt.BCryptDestroyKey(key_handle)


def _decrypt(cipher, key, data, iv, padding):
    """
    Decrypts AES/RC4/RC2/3DES/DES ciphertext

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    if not isinstance(key, byte_cls):
        raise TypeError(pretty_message(
            '''
            key must be a byte string, not %s
            ''',
            type_name(key)
        ))

    if not isinstance(data, byte_cls):
        raise TypeError(pretty_message(
            '''
            data must be a byte string, not %s
            ''',
            type_name(data)
        ))

    if cipher != 'rc4' and not isinstance(iv, byte_cls):
        raise TypeError(pretty_message(
            '''
            iv must be a byte string, not %s
            ''',
            type_name(iv)
        ))

    if cipher not in set(['rc4', 'aes']) and not padding:
        raise ValueError('padding must be specified')

    if _backend == 'winlegacy':
        return _advapi32_decrypt(cipher, key, data, iv, padding)
    return _bcrypt_decrypt(cipher, key, data, iv, padding)


def _advapi32_decrypt(cipher, key, data, iv, padding):
    """
    Decrypts AES/RC4/RC2/3DES/DES ciphertext via CryptoAPI

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    context_handle = None
    key_handle = None

    try:
        context_handle, key_handle = _advapi32_create_handles(cipher, key, iv)

        if cipher == 'aes' and not padding and len(data) % 16 != 0:
            raise ValueError('Invalid data - ciphertext length must be a multiple of 16')

        buffer = buffer_from_bytes(data)
        out_len = new(advapi32, 'DWORD *', len(data))
        res = advapi32.CryptDecrypt(
            key_handle,
            null(),
            # To skip padding, we have to tell the API that this is not
            # the final block
            False if cipher == 'aes' and not padding else True,
            0,
            buffer,
            out_len
        )
        handle_error(res)

        return bytes_from_buffer(buffer, deref(out_len))

    finally:
        if key_handle:
            advapi32.CryptDestroyKey(key_handle)
        if context_handle:
            close_context_handle(context_handle)


def _bcrypt_decrypt(cipher, key, data, iv, padding):
    """
    Decrypts AES/RC4/RC2/3DES/DES ciphertext via CNG

    :param cipher:
        A unicode string of "aes", "des", "tripledes_2key", "tripledes_3key",
        "rc2", "rc4"

    :param key:
        The encryption key - a byte string 5-16 bytes long

    :param data:
        The ciphertext - a byte string

    :param iv:
        The initialization vector - a byte string - unused for RC4

    :param padding:
        Boolean, if padding should be used - unused for RC4

    :raises:
        ValueError - when any of the parameters contain an invalid value
        TypeError - when any of the parameters are of the wrong type
        OSError - when an error is returned by the OS crypto library

    :return:
        A byte string of the plaintext
    """

    key_handle = None

    try:
        key_handle = _bcrypt_create_key_handle(cipher, key)

        if iv is None:
            iv_len = 0
        else:
            iv_len = len(iv)

        flags = 0
        if padding is True:
            flags = BcryptConst.BCRYPT_BLOCK_PADDING

        out_len = new(bcrypt, 'ULONG *')
        res = bcrypt.BCryptDecrypt(
            key_handle,
            data,
            len(data),
            null(),
            null(),
            0,
            null(),
            0,
            out_len,
            flags
        )
        handle_error(res)

        buffer_len = deref(out_len)
        buffer = buffer_from_bytes(buffer_len)
        iv_buffer = buffer_from_bytes(iv) if iv else null()

        res = bcrypt.BCryptDecrypt(
            key_handle,
            data,
            len(data),
            null(),
            iv_buffer,
            iv_len,
            buffer,
            buffer_len,
            out_len,
            flags
        )
        handle_error(res)

        return bytes_from_buffer(buffer, deref(out_len))

    finally:
        if key_handle:
            bcrypt.BCryptDestroyKey(key_handle)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import ctypes
from ctypes import windll, wintypes, POINTER, c_longlong, Structure

from .._ffi import FFIEngineError
from .._types import str_cls
from ..errors import LibraryNotFoundError


__all__ = [
    'get_error',
    'kernel32',
]


try:
    kernel32 = windll.kernel32
except (OSError) as e:
    if str_cls(e).find('The specified module could not be found') != -1:
        raise LibraryNotFoundError('kernel32.dll could not be found')
    raise

LARGE_INTEGER = c_longlong

try:
    kernel32.QueryPerformanceCounter.argtypes = [POINTER(LARGE_INTEGER)]
    kernel32.QueryPerformanceCounter.restype = wintypes.BOOL

    class FILETIME(Structure):
        _fields_ = [
            ("dwLowDateTime", wintypes.DWORD),
            ("dwHighDateTime", wintypes.DWORD),
        ]

    kernel32.GetSystemTimeAsFileTime.argtypes = [POINTER(FILETIME)]
    kernel32.GetSystemTimeAsFileTime.restype = None

except (AttributeError):
    raise FFIEngineError('Error initializing ctypes')


setattr(kernel32, 'LARGE_INTEGER', LARGE_INTEGER)
setattr(kernel32, 'FILETIME', FILETIME)


def get_error():
    error = ctypes.GetLastError()
    return (error, ctypes.FormatError(error))
